"""
A module for working with the RINKO analog optode.
"""

import logging
from collections import namedtuple
from pathlib import Path

import numpy as np
import pandas as pd
import scipy

from . import ctd_plots, flagging, get_ctdcal_config, oxy_fitting, process_ctd

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)

RinkoO2Cal = namedtuple("RinkoO2Cal", [*"ABCDEFGH"])
RinkoTMPCal = namedtuple("RinkoTMPCal", [*"ABCD"])


def rinko_DO(p_prime, G, H):
    """
    Calculates the dissolved oxygen percentage.
    """

    DO = G + H * p_prime

    return DO


def rinko_p_prime(N, t, A, B, C, D, E, F, G, H):
    """
    Per RinkoIII manual: 'The film sensing the water is affect by environment
    temperature and pressure at the depth where it is deployed. Based on experiments,
    an empirical algorithm as following is used to correct data dissolved oxygen.'

    Parameters
    ----------
    N : array-like
        Raw instrument output
    t : array-like
        Temperature [degC]
    A-H : float
        Calibration parameters
    """
    p_prime = A / (1 + D * (t - 25)) + B / ((N - F) * (1 + D * (t - 25)) + C + F)

    return p_prime


def correct_pressure(P, d, E):
    """
    Parameters
    ----------
    P : array-like
        Temperature-corrected DO [%]
    d : array-like
        Pressure [MPa]
    E : float
        Manufacturer calibration coefficient

    Returns
    -------
    P_d : array-like
        Temperature- and pressure-corrected DO [%]
    """
    # what is the dbar ~ MPa?

    P_d = P * (1 + E * d)

    return P_d


def salinity_correction(DO_c, T, S):
    """
    Oxygen optode is not able to detect salinity, so a correction is applied to
    account for the effect of salt on oxygen concentration. See Uchida (2010) in
    GO-SHIP manual (pg. 6, eq. 9) for more info.

    Parameters
    ----------
    DO_c : array-like
        Pressure-corrected dissolved oxygen
    T : array-like
        Calibrated CTD temperature
    S : array-like
        Calibrated CTD salinity

    Returns
    -------
    DO_sc : array-like
        Pressure- and salinity-corrected dissolved oxygen
    """
    # solubility coefficients from Benson and Krause (1984),
    # as recommended by Garcia and Gordon (1992)
    B0 = -6.24523e-3
    B1 = -7.37614e-3
    B2 = -1.03410e-2
    B3 = -8.17083e-3
    C0 = -4.88682e-7

    # "scaled temperature"
    T_scaled = np.log((298.15 - T) / (273.15 + T))

    # correction equation
    DO_sc = DO_c * np.exp(
        S * (B0 + (B1 * T_scaled) + (B2 * T_scaled ** 2) + (B3 * T_scaled ** 3))
        + C0 * S ** 2
    )

    return DO_sc


def _Uchida_DO_eq(coefs, inputs):
    """
    See Uchida et. al (2008) for more info:
    https://doi.org/10.1175/2008JTECHO549.1
    and Uchida et. al (2010) - GO-SHIP manual

    Parameters
    ----------
    coefs : tuple
        (c0, c1, c2, d0, d1, d2, cp)
    inputs : tuple
        (raw voltage, pressure, temperature, salinity, oxygen solubility)
    """
    c0, c1, c2, d0, d1, d2, cp = coefs
    V_r, P, T, S, o2_sol = inputs

    K_sv = c0 + (c1 * T) + (c2 * T ** 2)  # Stern-Volmer constant (Tengberg et al. 2006)
    V0 = (1 + d0 * T)  # voltage at zero oxygen (Uchida 2010, eq. 10)
    Vc = (d1 + d2 * V_r)  # raw voltage (Uchida 2010, eq. 10)
    o2_sat = ((V0 / Vc) - 1) / K_sv  # oxygen saturation [%] (Uchida 2010, eq. 6)

    DO = o2_sat * o2_sol  # dissolved oxygen concentration
    DO_c = DO * (1 + cp * P / 1000) ** (1 / 3)  # pressure compensated DO
    DO_sc = salinity_correction(DO_c, T, S)  # salinity + pressure compensated DO

    return DO_sc


def oxy_weighted_residual(coefs, weights, inputs, refoxy, L_norm=2):
    """
    A weighted residual fit in a similar manner to that of the SBE43 method of oxy_fitting.py.
    """
    # (abstracted from PMEL code oxygen_cal_ml.m)
    # unweighted L2: sum((ref - oxy)^2)  # if weighted fails
    # unweighted L4: sum((ref - oxy)^4)  # unsure of use case
    # unweighted L1: sum(abs(ref - oxy))  # very far from ideal
    # anything else? genericize with integer "norm" function input?

    residuals = np.sum(
        (weights * (refoxy - _Uchida_DO_eq(coefs, inputs)) ** 2)
    ) / np.sum(weights ** 2)

    return residuals


def calibrate_oxy(btl_df, time_df, ssscc_list):
    """
    Non-linear least squares fit oxygen optode against bottle oxygen.

    Note: optode data that were obtained during bottle stops can be used for calibration
    instead of density matching the downcast (see Uchida 2010, pg. 7).

    Parameters
    ----------
    btl_df : DataFrame
        CTD data at bottle stops
    time_df : DataFrame
        Continuous CTD data
    ssscc_list : list of str
        List of stations to process

    Returns
    -------
    """

    log.info("Calibrating oxygen (RINKO)")

    # initialize coef df
    coefs_df = pd.DataFrame(columns=["c0", "c1", "c2", "d0", "d1", "d2", "cp"])

    # Only fit using OXYGEN flagged good (2)
    good_data = btl_df[btl_df["OXYGEN_FLAG_W"] == 2].copy()

    # Fit ALL oxygen stations together to get initial coefficient guess
    (rinko_coefs0, _) = rinko_oxy_fit(good_data, f_suffix="_r0")
    coefs_df.loc["r0"] = rinko_coefs0  # log for comparison

    # fit station groups, like T/C fitting (ssscc_r1, _r2, etc.)
    ssscc_subsets = sorted(Path(cfg.dirs["ssscc"]).glob("ssscc_r*.csv"))
    if not ssscc_subsets:  # if no r-segments exists, write one from full list
        log.debug(
            "No CTDRINKO grouping file found... creating ssscc_r1.csv with all casts"
        )
        if not Path(cfg.dirs["ssscc"]).exists():
            Path(cfg.dirs["ssscc"]).mkdir()
        ssscc_list = process_ctd.get_ssscc_list()
        ssscc_subsets = [Path(cfg.dirs["ssscc"] + "ssscc_r1.csv")]
        pd.Series(ssscc_list).to_csv(ssscc_subsets[0], header=None, index=False)
    for f in ssscc_subsets:
        ssscc_sublist = pd.read_csv(f, header=None, dtype="str", comment='#').squeeze().to_list()
        f_stem = f.stem
        (rinko_coefs_group, _) = rinko_oxy_fit(
            good_data.loc[good_data["SSSCC"].isin(ssscc_sublist)].copy(),
            rinko_coef0=rinko_coefs0,
            f_suffix=f"_{f_stem}",
        )
        coefs_df.loc[f_stem.split("_")[1]] = rinko_coefs_group  # log for comparison

        # deal with time dependent coefs by further fitting individual casts
        # NOTE (4/9/21): tried adding time drift term unsuccessfully
        # Uchida (2010) says fitting individual stations is the same (even preferred?)
        for ssscc in ssscc_sublist:
            (rinko_coefs_ssscc, _) = rinko_oxy_fit(
                good_data.loc[good_data["SSSCC"] == ssscc].copy(),
                rinko_coef0=rinko_coefs_group,
                f_suffix=f"_{ssscc}",
            )

            # check mean/stdev to see if new fit is better or worse
            btl_rows = btl_df["SSSCC"] == ssscc
            time_rows = time_df["SSSCC"] == ssscc
            group_resid = (
                _Uchida_DO_eq(
                    rinko_coefs_group,
                    (
                        btl_df.loc[btl_rows, cfg.column["rinko_oxy"]],
                        btl_df.loc[btl_rows, cfg.column["p"]],
                        btl_df.loc[btl_rows, cfg.column["t1"]],
                        btl_df.loc[btl_rows, cfg.column["sal"]],
                        btl_df.loc[btl_rows, "OS"],
                    ),
                )
                - btl_df.loc[btl_rows, "OXYGEN"]
            )
            ssscc_resid = (
                _Uchida_DO_eq(
                    rinko_coefs_ssscc,
                    (
                        btl_df.loc[btl_rows, cfg.column["rinko_oxy"]],
                        btl_df.loc[btl_rows, cfg.column["p"]],
                        btl_df.loc[btl_rows, cfg.column["t1"]],
                        btl_df.loc[btl_rows, cfg.column["sal"]],
                        btl_df.loc[btl_rows, "OS"],
                    ),
                )
                - btl_df.loc[btl_rows, "OXYGEN"]
            )
            worse_mean = np.abs(ssscc_resid.mean()) > np.abs(group_resid.mean())
            worse_stdev = ssscc_resid.std() > group_resid.std()
            if worse_mean and worse_stdev:
                log.info(
                    f"{ssscc} fit parameters worse than {f_stem} group – reverting back"
                )
                rinko_coefs_ssscc = rinko_coefs_group

            # apply coefficients
            btl_df.loc[btl_rows, "CTDRINKO"] = _Uchida_DO_eq(
                rinko_coefs_ssscc,
                (
                    btl_df.loc[btl_rows, cfg.column["rinko_oxy"]],
                    btl_df.loc[btl_rows, cfg.column["p"]],
                    btl_df.loc[btl_rows, cfg.column["t1"]],
                    btl_df.loc[btl_rows, cfg.column["sal"]],
                    btl_df.loc[btl_rows, "OS"],
                ),
            )
            time_df.loc[time_rows, "CTDRINKO"] = _Uchida_DO_eq(
                rinko_coefs_ssscc,
                (
                    time_df.loc[time_rows, cfg.column["rinko_oxy"]],
                    time_df.loc[time_rows, cfg.column["p"]],
                    time_df.loc[time_rows, cfg.column["t1"]],
                    time_df.loc[time_rows, cfg.column["sal"]],
                    time_df.loc[time_rows, "OS"],
                ),
            )

            # save coefficients to dataframe
            coefs_df.loc[ssscc] = rinko_coefs_ssscc

    # flag CTDRINKO with more than 1% difference
    time_df["CTDRINKO_FLAG_W"] = 2
    btl_df["CTDRINKO_FLAG_W"] = flagging.by_percent_diff(
        btl_df["CTDRINKO"], btl_df["OXYGEN"], percent_thresh=1
    )

    # Plot all post fit data
    f_out = f"{cfg.fig_dirs['rinko']}rinko_residual_all_postfit.pdf"
    ctd_plots._intermediate_residual_plot(
        btl_df["OXYGEN"] - btl_df["CTDRINKO"],
        btl_df["CTDPRS"],
        btl_df["SSSCC"],
        xlabel="CTDRINKO Residual (umol/kg)",
        f_out=f_out,
        xlim=(-10, 10),
    )
    f_out = f"{cfg.fig_dirs['rinko']}rinko_residual_all_postfit_flag2.pdf"
    flag2 = btl_df["CTDRINKO_FLAG_W"] == 2
    ctd_plots._intermediate_residual_plot(
        btl_df.loc[flag2, "OXYGEN"] - btl_df.loc[flag2, "CTDRINKO"],
        btl_df.loc[flag2, "CTDPRS"],
        btl_df.loc[flag2, "SSSCC"],
        xlabel="CTDRINKO Residual (umol/kg)",
        f_out=f_out,
        xlim=(-10, 10),
    )

    # export fitting coefs
    coefs_df.applymap(
        lambda x: np.format_float_scientific(x, precision=4, exp_digits=1)
    ).to_csv(cfg.dirs["logs"] + "rinko_coefs.csv")

    return True


def rinko_oxy_fit(
    btl_df,
    rinko_coef0=(
        1.89890,
        1.71137e-2,
        1.59838e-4,
        -1.07941e-3,
        -1.23152e-1,
        3.06114e-1,
        4.50828e-2,
    ),
    f_suffix=None,
):
    """
    Iteratively fit Rinko DO data against bottle oxygen.

    Default coefficients come from an old cruise report:
    https://cchdo.ucsd.edu/data/2362/p09_49RY20100706do.txt
    (there's probably a better way – are there physical meanings?)
    """
    # # Plot data to be fit together
    # f_out = f"{cfg.fig_dirs['ox']}rinko_residual{f_suffix}_prefit.pdf"
    # ctd_plots._intermediate_residual_plot(
    #     merged_df["REFOXY"] - merged_df[cfg.column["rinko_oxy"]],
    #     merged_df["CTDPRS"],
    #     merged_df["SSSCC"],
    #     xlabel="CTDRINKO Residual (umol/kg)",
    #     f_out=f_out,
    #     xlim=(-10, 10),
    # )

    bad_df = pd.DataFrame()
    weights = oxy_fitting.calculate_weights(btl_df["CTDPRS"])
    fit_data = (
        btl_df[cfg.column["rinko_oxy"]],
        btl_df[cfg.column["p"]],
        btl_df[cfg.column["t1"]],
        btl_df[cfg.column["sal"]],
        btl_df["OS"],
    )
    # bounds need to be specified for all, can't just do cp...
    coef_bounds = [
        (None, None),  # c0,
        (None, None),  # c1,
        (None, None),  # c2,
        (None, None),  # d0,
        (None, None),  # d1,
        (None, None),  # d2,
        (0, 0.2),  # cp, pressure compensation
    ]
    res = scipy.optimize.minimize(
        oxy_weighted_residual,
        x0=rinko_coef0,
        args=(weights, fit_data, btl_df[cfg.column["refO"]]),
        bounds=coef_bounds,
    )

    cfw_coefs = res.x
    btl_df["RINKO_OXY"] = _Uchida_DO_eq(cfw_coefs, fit_data)
    btl_df["residual"] = btl_df[cfg.column["refO"]] - btl_df["RINKO_OXY"]

    cutoff = 2.8 * np.std(btl_df["residual"])
    thrown_values = btl_df[np.abs(btl_df["residual"]) > cutoff]
    bad_df = pd.concat([bad_df, thrown_values])
    btl_df = btl_df[np.abs(btl_df["residual"]) <= cutoff].copy()

    while not thrown_values.empty:

        p0 = tuple(cfw_coefs)
        weights = oxy_fitting.calculate_weights(btl_df["CTDPRS"])
        fit_data = (
            btl_df[cfg.column["rinko_oxy"]],
            btl_df[cfg.column["p"]],
            btl_df[cfg.column["t1"]],
            btl_df[cfg.column["sal"]],
            btl_df["OS"],
        )
        res = scipy.optimize.minimize(
            oxy_weighted_residual,
            x0=p0,
            args=(weights, fit_data, btl_df[cfg.column["refO"]]),
            bounds=coef_bounds,
        )
        cfw_coefs = res.x
        btl_df["RINKO_OXY"] = _Uchida_DO_eq(cfw_coefs, fit_data)
        btl_df["residual"] = btl_df[cfg.column["refO"]] - btl_df["RINKO_OXY"]
        cutoff = 2.8 * np.std(btl_df["residual"])
        thrown_values = btl_df[np.abs(btl_df["residual"]) > cutoff]
        bad_df = pd.concat([bad_df, thrown_values])
        btl_df = btl_df[np.abs(btl_df["residual"]) <= cutoff].copy()

    # intermediate plots to diagnose data chunks goodness
    if f_suffix is not None:
        f_out = f"{cfg.fig_dirs['rinko']}rinko_residual{f_suffix}.pdf"
        ctd_plots._intermediate_residual_plot(
            btl_df["residual"],
            btl_df["CTDPRS"],
            btl_df["SSSCC"],
            xlabel="CTDRINKO Residual (umol/kg)",
            f_out=f_out,
            xlim=(-10, 10),
        )

    btl_df["CTDRINKO_FLAG_W"] = 2
    bad_df["CTDRINKO_FLAG_W"] = 3
    df = pd.concat([btl_df, bad_df])

    return cfw_coefs, df

    # from . import ctd_plots
    # diff = all_rinko_merged["REFOXY"] - all_rinko_merged["RINKO_OXY"]
    # rows = all_rinko_merged["SSSCC"] == "00101"
    # plt.scatter(all_rinko_merged["REFOXY"], all_rinko_merged["CTDPRS"])
    # plt.scatter(all_rinko_merged["RINKO_OXY"], all_rinko_merged["CTDPRS"])
    # ctd_plots._intermediate_residual_plot(diff, all_rinko_merged["CTDPRS"], all_rinko_merged["SSSCC"], xlim=(-10,10), f_out="rinko_test_residual.pdf")


### Everything below is old code (not necessarily bad)



def rinko_o2_cal_parameters(**kwargs):
    """
    Calibration coefficients for the oxygen calculations (from RinkoIII manual). WARNING: Requires manual update for new cal. coeffs.
    
    Note: These are SN specific and are likely outdated.
    """
    A = kwargs.get("A", -4.524084e1)
    B = kwargs.get("B",  1.449377e2)
    C = kwargs.get("C", -3.051590e-1)
    D = kwargs.get("D",  1.065300e-2)
    E = kwargs.get("E",  4.000000e-3)
    F = kwargs.get("F",  6.250000e-5)
    G = kwargs.get("G",  0.000000e0)
    H = kwargs.get("H",  1.000000e0)
    return RinkoO2Cal(A,B,C,D,E,F,G,H)

def rinko_temperature_cal_parameters(**kwargs):
    """
    Rinko temperature calibration coefficients (from RinkoIII manual). WARNING: Requires manual update for new cal. coeffs.

    Note: These are SN specific and are likely outdated.
    """
    A = kwargs.get("A", -5.305905e0)
    B = kwargs.get("B",  1.666857e1)
    C = kwargs.get("C", -2.142681e0)
    D = kwargs.get("D",  4.582805e-1)
    return RinkoTMPCal(A,B,C,D)

def rinko_temperature(v, tmp_cal:RinkoTMPCal):
    """
    Calculate rinko temperature from voltage and calibration coeffieicnets.
    """
    if type(tmp_cal) is not RinkoTMPCal:
        raise ValueError("tmp_cal must be of type RinkoTMPCal")

    A, B, C, D = tmp_cal
    return A + B*v + C*v**2 + D*v**3

def rinko_pprime_aro_cav(v, t, o2_cal:RinkoO2Cal):
    """
    Calculates Rinko P' of the equation P = G + H * P'
    where P is DO physical value IN PERCENT [%]
    """
    A, B, C, D, E, F, G, H = o2_cal

    term_1_denominator = 1 + D*(t-25) + F*(t-25)**2
    term_1 = A/term_1_denominator

    term_2_denominator = v * (1 + D*(t-25) + F*(t-25)**2) + C
    term_2 = B/term_2_denominator

    return term_1 + term_2

def rinko_saturation(pprime, o2_cal:RinkoO2Cal):
    """
        Calculates Rinko P of the equation P = G + H * P'
        where P is DO physical value IN PERCENT [%]
    """

    A, B, C, D, E, F, G, H = o2_cal

    return G + H * pprime

def rinko_correct_for_pressure(p, d, o2_cal:RinkoO2Cal):
    """Note that the pressure term, d, must be in MPa

    1 decibar = 0.01 Mpa
    """
    A, B, C, D, E, F, G, H = o2_cal

    return p*(1 + E*d)

#   20240612 AJM commenting out method, possibly redundant
# def rinko_saturation(df, film="B", model="ARO-CAV", **kwargs):
#     """
#     Derive oxygen saturation for the RINKO. Unused as of 2021.
#     """
#     pass

def rinko_oxy_eq(press, temp, oxyvo, os, o2_cal:RinkoO2Cal):
    """
    Derive corrected RINKO dissolved oxygen.
    """
    #Calculate pprime

    pprime = rinko_pprime_aro_cav(oxyvo,temp,o2_cal)

    # Calculate P (DO physical value in %)

    #p = rinko_saturation(pprime, o2_cal)

    # Correct for pressure * d is pressure in Mpa *

    d = press * 0.01

    p_corr = rinko_correct_for_pressure(pprime,d,o2_cal)

    # Divide by 100 to get percents in the form of 0.xx

    p_corr = p_corr / 100

    # Multiply by OS to get DO (os can be in either ml/l or umol/kg)

    DO = p_corr * os

    return DO

def rinko_curve_fit_eq(X, a, b, c, d, e, f, g, h):
    """
    Same as rinko_oxy_eq, but in a form that is more suitible for scipy's curve fit routine
    X contains pressure, temperature, voltage, and OS (the normal arguments for rinko_oxy_eq)
    """

    press, temp, oxyvo, os = X
    o2_cal = RinkoO2Cal(a, b, c, d, e, f, g, h)

    #Calculate pprime

    pprime = rinko_pprime_aro_cav(oxyvo,temp,o2_cal)

    # Calculate P (DO physical value in %)

    #p = rinko_saturation(pprime, o2_cal)

    # Correct for pressure * d is pressure in Mpa *

    d = press * 0.01

    p_corr = rinko_correct_for_pressure(pprime,d,o2_cal)

    # Divide by 100 to get percents in the form of 0.xx

    p_corr = p_corr / 100

    # Multiply by OS to get DO (os can be in either ml/l or umol/kg)

    DO = p_corr * os

    return DO

def rinko_oxygen_cal(o2_cal,pressure,temp,oxyvolts,os,ref_oxy,switch):
    """"
    Rinko oxygen fitting routine using the equation:
        Calculates Rinko P of the equation P = G + H * P'
        where P is DO physical value IN PERCENT [%]

    Fits 7 Coef:
    coef[0] = A
    coef[1] = B
    coef[2] = C
    coef[3] = D
    coef[4] = E
    coef[5] = F
    coef[6] = G
    """

    rinko_oxy = rinko_oxy_eq(pressure, temp, oxyvolts, os, o2_cal)

    #Weight Determination
    if switch == 1:

        weights = oxy_fitting.calculate_weights(pressure)

        resid = ((weights * (ref_oxy - rinko_oxy))**2) / (np.sum(weights)**2) #Original way (np.sum(weights)**2)

    elif switch == 2:
        #L2 Norm
        resid = (ref_oxy - rinko_oxy)**2

    elif switch == 3:
        #ODF residuals

        resid = np.sqrt(((ref_oxy - rinko_oxy)**2) / (np.std(rinko_oxy)**2))

    elif switch == 4:
        # Weighted ODF residuals

        weights = oxy_fitting.calculate_weights(pressure)
        resid = np.sqrt(weights * ((ref_oxy - rinko_oxy)**2) / (np.sum(weights)**2))#(np.std(ctd_oxy_mlL)**2))

    elif switch == 5:

        weights = oxy_fitting.calculate_weights(pressure)

        resid = ((weights * (ref_oxy - rinko_oxy))**2) / (np.sum(weights**2))


    return resid

def rinko_weighted_residual(coefs,weights,inputs,refoxy):
    """
    Generate weights from oxygen residuals and provided coefficients.
    """
    a, b, c, d, e, f, g, h = coefs
    return np.sum((weights*(refoxy-rinko_curve_fit_eq(inputs, a, b, c, d, e, f, g, h))**2))/np.sum(weights**2)

def match_sigmas(btl_prs, btl_oxy, btl_sigma, ctd_sigma, ctd_os, ctd_prs, ctd_tmp, rinkovolts, btl_ssscc=None):
    """
    Density matching in a manner similar to that of the SBE43 method described in oxy_fitting.py.
    """
    # Construct Dataframe from bottle and ctd values for merging
    if 'btl_ssscc' in locals():
        btl_dict = {'CTDPRS_rinko_btl':btl_prs, 'REFOXY_rinko':btl_oxy, 'sigma_rinko_btl':btl_sigma, 'SSSCC_rinko':btl_ssscc}
    else:
        btl_dict = {'CTDPRS_rinko_btl':btl_prs, 'REFOXY_rinko':btl_oxy, 'sigma_rinko_btl':btl_sigma}
    btl_data = pd.DataFrame(btl_dict)
    time_dict = {'CTDPRS_rinko_ctd':ctd_prs, 'sigma_rinko_ctd':ctd_sigma, 'OS_rinko_ctd':ctd_os, 'CTDTMP_rinko':ctd_tmp, 'CTDRINKOVOLTS':rinkovolts}
    time_data = pd.DataFrame(time_dict)

    # Sort DataFrames by sigma0
    time_data.sort_values('sigma_rinko_ctd', inplace=True)
    btl_data.sort_values('sigma_rinko_btl', inplace=True)
    btl_data.dropna(subset=['REFOXY_rinko'], inplace=True)

    # Merge DF
    merged_df = pd.merge_asof(btl_data, time_data, left_on='sigma_rinko_btl', right_on='sigma_rinko_ctd', direction='nearest', suffixes=['_btl','_ctd'])

    # Apply coef and calculate CTDRINKO
    rinko_coef0 = rinko_o2_cal_parameters()
    merged_df['CTDRINKO'] = rinko_oxy_eq(merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd'],rinko_coef0)

    return merged_df

#   20240612 AJM commenting out, as large portions are already commented out and is in non-functional state.    
# def rinko_oxygen_fit(merged_df, rinko_coef0=None, f_out=None):
#     """
#     An old method for RINKO fitting, replaced by calibrate_oxy. Similar to that of the SBE43.
#     """
#     # Create DF for good and questionable values

#     bad_df = pd.DataFrame()
#     good_df = pd.DataFrame()

#     if rinko_coef0 is None:
#         # Load initial coefficient guess
#         rinko_coef0 = rinko_o2_cal_parameters()

#     p0 = [x for x in rinko_coef0]

#     # Curve fit (weighted)
#     weights = oxy_fitting.calculate_weights(merged_df['CTDPRS_rinko_ctd'])
#     cfw_coefs = scipy.optimize.fmin(
#         rinko_weighted_residual,
#         x0=p0,
#         args=(
#             weights,(
#                 merged_df['CTDPRS_rinko_ctd'],
#                 merged_df['CTDTMP_rinko'],
#                 merged_df['CTDRINKOVOLTS'], 
#                 merged_df['OS_rinko_ctd']
#             ),
#             merged_df['REFOXY_rinko'],
#         ),
#         maxfun=10000,
#     )
#     merged_df['CTDRINKO'] = rinko_oxy_eq(merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd'],cfw_coefs)
    
#     merged_df['res_rinko'] = merged_df['REFOXY_rinko'] - merged_df['CTDRINKO']
#     stdres = np.std(merged_df['res_rinko'])
#     cutoff = stdres * 2.8

#     thrown_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
#     bad_df = pd.concat([bad_df, thrown_values])
#     merged_df = merged_df[np.abs(merged_df['res_rinko']) <= cutoff]

#     while not thrown_values.empty:  # runs as long as there are thrown_values

#         p0 = cfw_coefs
#         weights = oxy_fitting.calculate_weights(merged_df['CTDPRS_rinko_ctd'])
#         cfw_coefs = scipy.optimize.fmin(
#             rinko_weighted_residual,
#             x0=p0,
#             args=(
#                 weights,(
#                     merged_df['CTDPRS_rinko_ctd'],
#                     merged_df['CTDTMP_rinko'],
#                     merged_df['CTDRINKOVOLTS'], 
#                     merged_df['OS_rinko_ctd']
#                 ),
#                 merged_df['REFOXY_rinko'],
#             ),
#             maxfun=10000,
#         )
#         merged_df['CTDRINKO'] = rinko_oxy_eq(merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd'],cfw_coefs)

#         merged_df['res_rinko'] = merged_df['REFOXY_rinko'] - merged_df['CTDRINKO']
#         stdres = np.std(merged_df['res_rinko'])
#         cutoff = stdres * 2.8
#         thrown_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
#         print(len(thrown_values))
#         print(p0)
#         bad_df = pd.concat([bad_df, thrown_values])
#         merged_df = merged_df[np.abs(merged_df['res_rinko']) <= cutoff]

    # try:
    #     cfw_coef , cov = scipy.optimize.curve_fit(rinko_curve_fit_eq,
    #                                               (merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd']),
    #                                               merged_df['REFOXY_rinko'], coef0, sigma=weights, absolute_sigma=False, maxfev=50000)

    #     merged_df['CTDRINKO'] = rinko_oxy_eq(merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd'],cfw_coef)
    #     merged_df['res_rinko'] = merged_df['REFOXY_rinko'] - merged_df['CTDRINKO']
    #     stdres = np.std(merged_df['res_rinko'])
    #     cutoff = stdres * 2.8

    #     thrown_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #     bad_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #     bad_df = pd.concat([bad_df, bad_values])
    #     merged_df = merged_df[np.abs(merged_df['res_rinko']) <= cutoff]


    #     while not thrown_values.empty:

    #         p0 = cfw_coef
    #         # weights = 1/(np.power(merged_df['CTDPRS_rinko_ctd'],1/2))
    #         weights = 1/(oxy_fitting.calculate_weights(merged_df['CTDPRS_rinko_ctd']))
    #         cfw_coef , cov = scipy.optimize.curve_fit(rinko_curve_fit_eq, (merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd']), merged_df['REFOXY_rinko'], p0, sigma=weights, absolute_sigma=False, maxfev=50000)
    #         merged_df['CTDRINKO'] = rinko_oxy_eq(merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd'],cfw_coef)
    #         merged_df['res_rinko'] = merged_df['REFOXY_rinko'] - merged_df['CTDRINKO']
    #         stdres = np.std(merged_df['res_rinko'])
    #         cutoff = stdres * 2.8
    #         thrown_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #         bad_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #         bad_df = pd.concat([bad_df, bad_values])
    #         merged_df = merged_df[np.abs(merged_df['res_rinko']) <= cutoff]

    # except RuntimeError:

    #         try:#Nested try/except could be better
    #             print('Weighted curve fitting failed for RINKO...using Unweighted Fitting')
    #             cfw_coef , cov = scipy.optimize.curve_fit(rinko_curve_fit_eq, (merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd']), merged_df['REFOXY_rinko'], coef0)
    #             merged_df['CTDRINKO'] = rinko_oxy_eq(merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd'],cfw_coef)

    #             merged_df['res_rinko'] = merged_df['REFOXY_rinko'] - merged_df['CTDRINKO']
    #             stdres = np.std(merged_df['res_rinko'])
    #             cutoff = stdres * 2.8

    #             thrown_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #             bad_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #             bad_df = pd.concat([bad_df, bad_values])
    #             merged_df = merged_df[np.abs(merged_df['res_rinko']) <= cutoff]

    #             while not thrown_values.empty:

    #                 p0 = cfw_coef
    #                 cfw_coef , cov = scipy.optimize.curve_fit(rinko_curve_fit_eq, (merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd']), merged_df['REFOXY_rinko'], coef0)
    #                 merged_df['CTDRINKO'] = rinko_oxy_eq(merged_df['CTDPRS_rinko_ctd'],merged_df['CTDTMP_rinko'],merged_df['CTDRINKOVOLTS'],merged_df['OS_rinko_ctd'],cfw_coef)

    #                 merged_df['res_rinko'] = merged_df['REFOXY_rinko'] - merged_df['CTDRINKO']
    #                 stdres = np.std(merged_df['res_rinko'])
    #                 cutoff = stdres * 2.8

    #                 thrown_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #                 bad_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #                 bad_df = pd.concat([bad_df, bad_values])
    #                 merged_df = merged_df[np.abs(merged_df['res_rinko']) <= cutoff]

    #         except:
    #             print('Logging RINKO coef...')
    #             cfw_coef = coef0
    #             merged_df['res_rinko'] = merged_df['REFOXY_rinko'] - merged_df['CTDRINKO']
    #             stdres = np.std(merged_df['res_rinko'])
    #             cutoff = stdres * 2.8
    #             thrown_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #             bad_values = merged_df[np.abs(merged_df['res_rinko']) > cutoff]
    #             merged_df = merged_df[np.abs(merged_df['res_rinko']) <= cutoff]


    #coef_dict[station] = cfw_coef
    # good_df = pd.concat([good_df, merged_df])
    # good_df['CTDRINKO_FLAG_W'] = 2
    # bad_df = pd.concat([bad_df, bad_values])
    # bad_df['CTDRINKO_FLAG_W'] = 3
    # df = pd.concat([good_df,bad_df])
    # df.sort_values(by='CTDPRS_rinko_btl',ascending=False,inplace=True)
    # oxy_df = df.copy()

    # return cfw_coef, df


#   20240612 AJM commenting the below out, as it is not useful in the modern workflow
# if __name__ == "__main__":
#     print(rinko_temperature(1.565323565, rinko_temperature_cal_parameters()))
#     print(rinko_correct_for_pressure(
#     rinko_pprime_aro_cav(1.5421, 0.442, rinko_o2_cal_parameters()),
#     2/100,
#     rinko_o2_cal_parameters()
#     ))



import numpy as np
import pytest

from ctdcal import oxy_fitting, rinko

#   Coefs from CTDCAL run
coefs = (2.4837e+0, 1.7548e-1, -1.0020e-3, 3.1672e-2, -1.8098e-1, 2.685e-1, 9.7101e-2)
#   Fake pressure points
pressure = np.array([50, 150, 250, 400, 800, 1500, 2500, 5000])
refoxy = np.array([400, 160, 170, 200, 280, 335, 360, 400])
#   Fake inputs tuple
#   (raw voltage, pressure, temperature, salinity, oxygen solubility)
inputs = (1.5, 500, 20, 35, 100)


def test_rinko_DO():
    #   Assign test cases
    p_prime = 0.5
    G = 10
    H = 5

    assert rinko.rinko_DO(p_prime, G, H) == 10 + 5 * 0.5

def test_rinko_p_prime():
    #   Shared raw and temperature arrays for test cases
    #   Assuming "raw" are voltages
    N = np.array([1.7, 1.9, 2.0])
    t = np.array([20, 25, 22])

    A, B, C, D, E, F, G, H = 2, 3, 4, 5, 6, 7, 8, 9
    p_solved = (A / (1 + D * (t - 25))) + (B / ((N - F) * (1 + D * (t - 25)) + C + F))
    p_prime = rinko.rinko_p_prime(N, t, A, B, C, D, E, F, G, H)
    assert np.allclose(p_prime, p_solved)

    #   Now with 0297 coeffs
    A, B, C, D, E, F, G, H = -4.367428e+01, 1.376636e+02, -3.647983e-01, 1.044300e-02, 4.300000e-03, 6.810000e-05, 0, 1
    p_solved = (A / (1 + D * (t - 25))) + (B / ((N - F) * (1 + D * (t - 25)) + C + F))
    p_prime = rinko.rinko_p_prime(N, t, A, B, C, D, E, F, G, H)
    assert np.allclose(p_prime, p_solved)

def test_correct_pressure():
    #   Assign test cases
    P = np.array([80, 90, 85])  #   Corrected % oxygen
    d = np.array([0.1, 0.2, 0.15])  #   Assuming 1 MPa = 10 dbar
    E = 0.0043  #   Grabbed from SN 0297
    p_solved = P * (1 + E * d)

    p_corr = rinko.correct_pressure(P, d, E)

    assert np.allclose(p_corr, p_solved)

def test_salinity_correction():
    #   Assign test cases
    DO_c = np.array([400, 200, 180])
    T = np.array([20.5, 29.5, 33.0])
    S = np.array([33.0, 34.5, 35.2])
    B0 = -6.24523e-3
    B1 = -7.37614e-3
    B2 = -1.03410e-2
    B3 = -8.17083e-3
    C0 = -4.88682e-7
    T_scaled = np.log((298.15 - T) / (273.15 + T))
    DO_sc_manual = DO_c * np.exp(
        S * (B0 + (B1 * T_scaled) + (B2 * T_scaled ** 2) + (B3 * T_scaled ** 3))
        + C0 * S ** 2
    )

    DO_sc = rinko.salinity_correction(DO_c, T, S)
    assert np.allclose(DO_sc, DO_sc_manual)

    #   NaN handling
    S = np.array([35.0, np.nan, 34.5])
    DO_sc = rinko.salinity_correction(DO_c, T, S)
    assert np.isnan(DO_sc[1])

@pytest.mark.parametrize("coefs, inputs", [(coefs, inputs)])
def test_Uchida_DO_eq(coefs, inputs):
    V_r, P, T, S, o2_sol = inputs
    K_sv = coefs[0] + (coefs[1] * T) + (coefs[2] * T ** 2)
    V0 = (1 + coefs[3] * T)
    Vc = (coefs[4] + coefs[5] * V_r)
    o2_sat = ((V0 / Vc) - 1) / K_sv
    DO = o2_sat * o2_sol
    DO_c = DO * (1 + coefs[6] * P / 1000) ** (1 / 3)
    DO_sc_manual = rinko.salinity_correction(DO_c, T, S)

    DO_sc = rinko._Uchida_DO_eq(coefs, inputs)
    assert np.allclose(DO_sc, DO_sc_manual)

@pytest.mark.parametrize("coefs, pressure, inputs, refoxy", [(coefs, pressure, inputs, refoxy)])
def oxy_weighted_residual(coefs, weights, inputs, refoxy, pressure):

    weights = oxy_fitting.calculate_weights(pressure)
    residuals_manual = np.sum(
        (weights * (refoxy - rinko._Uchida_DO_eq(coefs, inputs)) ** 2)
    ) / np.sum(weights ** 2)

    residuals = oxy_weighted_residual(coefs, weights, inputs, refoxy, pressure)

    assert np.allclose(residuals, residuals_manual)


"""
A module for handling SeaBird raw .HEX files, including the generation of bottle-extractions, downcast isolation, and SBE3/4C handling.
"""

import logging
from pathlib import Path

import gsw
import numpy as np
import pandas as pd

from . import equations_sbe as sbe_eq
from . import get_ctdcal_config
from . import process_bottle as btl
from . import process_ctd as process_ctd
from . import sbe_reader as sbe_rd
from ctdcal.processors.cast_tools import Cast
from .common import validate_dir

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)

# lookup table for sensor data
# DOUBLE CHECK TYPE IS CORRECT #
short_lookup = {
    "55": {
        "short_name": "CTDTMP",
        "long_name": "SBE 3+ Temperature",
        "units": "ITS-90",
        "type": "float64",
    },
    "45": {
        "short_name": "CTDPRS",
        "long_name": "SBE 9+ Pressure",
        "units": "DBAR",
        "type": "float64",
    },
    "3": {
        "short_name": "CTDCOND",
        "long_name": "SBE 4 Conductivity",
        "units": "MSPCM",
        "type": "float64",
    },
    "38": {
        "short_name": "CTDOXY",
        "long_name": "SBE 43 Oxygen",
        "units": "MLPL",
        "type": "float64",
    },
    # '38':{'short_name': 'CTDOXYVOLTS', 'long_name':'SBE 43 Oxygen Volts', 'units': '0-5VDC', 'type':'float64'},
    "11": {
        "short_name": "FLUOR",
        "long_name": "Seapoint Fluorometer",
        "units": "0-5VDC",
        "type": "float64",
    },
    "27": {"short_name": "FREE", "long_name": "empty", "units": "NA", "type": "NA"},
    "0": {
        "short_name": "ALT",
        "long_name": "Altitude",
        "units": "M",
        "type": "float64",
    },
    "71": {
        "short_name": "CTDXMISS",
        "long_name": "CStar",
        "units": "0-5VDC",
        "type": "float64",
    },
    "61": {
        "short_name": "U_DEF_poly",
        "long_name": "user defined, polynomial",
        "units": "0-5VDC",
        "type": "float64",
    },
    "80": {
        "short_name": "U_DEF_e",
        "long_name": "user defined, exponential",
        "units": "0-5VDC",
        "type": "float64",
    },
    "1000": {
        "short_name": "CTDSAL",
        "long_name": "Salinity (C1 T1)",
        "units": "PSU",
        "type": "float64",
    },
    "20": {
        "short_name": "CTDFLUOR",
        "long_name": "WetlabECO_AFL_FL_Sensor",
        "units": "0-5VDC",
        "type": "float64",
    },
    "42": {
        "short_name": "PAR",
        "long_name": "PAR/Irradiance, Biospherical/Licor",
        "units": "0-5VDC",
        "type": "float64",
    },
    "51": {
        "short_name": "REF_PAR",
        "long_name": "Surface PAR/Irradiance, Biospherical/Licor",
        "units": "0-5VDC",
        "type": "float64",
    },
    "70": {
        "short_name": "CTDBACKSCATTER",
        "long_name": "WetlabECO_BB_Sensor",
        "units": "0-5VDC",
        "type": "float64",
    },
    "13": {
        "short_name": "CTDFLUOR",
        "long_name": "FluoroSeatechWetlabsFLF_Sensor",
        "units": "0-5VDC",
        "type": "float64",
    },
    "67": {
        "short_name": "TURBIDITY",
        "long_name": "Turbidity Meter",
        "units": "0-5VDC",
        "type": "float64",
    },
    "19": {
        "short_name": "FLUOR_CDOM",
        "long_name": "FluoroWetlabCDOM_Sensor",
        "units": "0-5VDC",
        "type": "float64",
    },
}


def hex_to_ctd(ssscc_list):
    """
    Convert raw CTD data and export to .pkl files.

    Parameters
    ----------
    ssscc_list : list of str
        List of stations to convert

    Returns
    -------

    """
    log.info("Converting .hex files")
    for ssscc in ssscc_list:
        if not Path(cfg.dirs["converted"] + ssscc + ".pkl").exists():
            hexFile = cfg.dirs["raw"] + ssscc + ".hex"
            xmlconFile = cfg.dirs["raw"] + ssscc + ".XMLCON"
            sbeReader = sbe_rd.SBEReader.from_paths(hexFile, xmlconFile)
            converted_df = convertFromSBEReader(sbeReader, ssscc)
            converted_df.to_pickle(cfg.dirs["converted"] + ssscc + ".pkl")

    return True


def make_time_files(casts, datadir, user_cfg):
    """
    Make continuous time-series files from processed cast data.

    Each cast has a smoothing filter applied. Filter parameters and columns
    to filter are from user-specified configurations.

    The time on deck, the soak, and the upcast are trimmed to provide a continuous downcast.
    
    Parameters
    ----------
    casts : list of str
        List of cast ids to process.
    datadir : str or Path-like
        Top-level of user data directory
    user_cfg : Munch object
        Dictionary of user configuration parameters.
    """
    log.info("Generating time.pkl files")
    # validate time directory
    time_dir = validate_dir(Path(datadir, 'time'), create=True)
    # groundwork for writing any new details or offsets
    details_file = Path(datadir, 'logs/cast_details.csv')
    offsets_file = Path(datadir, 'logs/ondeck_pressure.csv')
    new_casts = False
    if details_file.exists():
        cast_details_all = pd.read_csv(details_file, dtype='str')
    else:
        cast_details_all = pd.DataFrame()
    if offsets_file.exists():
        p_offsets_all = pd.read_csv(offsets_file, dtype='str')
    else:
        p_offsets_all = pd.DataFrame()

    # process new casts one by one
    for cast_id in casts:
        time_file = Path(time_dir, '%s_time.pkl' % cast_id)
        if not time_file.exists():
            new_casts = True
            cast = Cast(cast_id, datadir)
            cast.p_col = 'CTDPRS'
            # Apply smoothing filter
            cast.filter(cast.proc,
                        win_size=(user_cfg.filter_win * user_cfg.freq),
                        win_type=user_cfg.filter_type,
                        cols=user_cfg.filter_cols)
            # Parse the downcast from the full cast
            cast.parse_downcast(cast.filtered)
            # Trim the soak period from the downcast
            cast.trim_soak(cast.downcast,
                           (user_cfg.soak_win * user_cfg.freq),
                           user_cfg.soak_threshold)
            # save pkl file
            cast.trimmed.to_pickle(time_file)

            # AS: 2024-09-05 - leaving this here for reference. Despiking is currently
            # TBD for cast_tools post processing...
            #
            # Remove any pressure spikes
            # bad_rows = converted_df["CTDPRS"].abs() > 6500
            # if bad_rows.any():
            #     log.debug(f"{cast_id}: {bad_rows.sum()} bad pressure points removed.")
            # converted_df.loc[bad_rows, :] = np.nan
            # converted_df.interpolate(limit=24, limit_area="inside", inplace=True)

            # Merge in new details and offsets values
            cast_details_all = (pd.concat([cast_details_all, cast.get_details()])
                                .drop_duplicates(['cast_id'], keep='last')
                                .sort_values(by=['cast_id']))
            p_offsets_all = (pd.concat([p_offsets_all,
                                       cast.get_pressure_offsets(cast.proc,
                                                                 user_cfg.cond_threshold,
                                                                 user_cfg.freq)])
                             .drop_duplicates(['cast_id'], keep='last')
                             .sort_values(by=['cast_id']))

    # Wrap up...
    if new_casts is True:
        log.info("Saving deck pressures and cast details.")
        cast_details_all.to_csv(Path(datadir, 'logs/cast_details.csv'), index=False)
        p_offsets_all.to_csv(Path(datadir, 'logs/ondeck_pressure.csv'), index=False)

def make_btl_mean(ssscc_list):
    """
    Create "bottle mean" files from continuous CTD data averaged at the bottle stops.

    Parameters
    ----------
    ssscc_list : list of str
        List of stations to convert

    Returns
    -------
    boolean
        bottle averaging of mean has finished successfully
    """
    log.info("Generating btl_mean.pkl files")
    for ssscc in ssscc_list:
        if not Path(cfg.dirs["bottle"] + ssscc + "_btl_mean.pkl").exists():
            imported_df = pd.read_pickle(cfg.dirs["converted"] + ssscc + ".pkl")
            bottle_df = btl.retrieveBottleData(imported_df)
            mean_df = btl.bottle_mean(bottle_df)

            # export bottom bottle time/lat/lon info
            fname = cfg.dirs["logs"] + "bottom_bottle_details.csv"
            datetime_col = "nmea_datetime"
            if datetime_col not in mean_df.columns:
                log.debug(
                    f"'{datetime_col}' not found in DataFrame - using 'scan_datetime'"
                )
                datetime_col = "scan_datetime"

            bot_df = mean_df[[datetime_col, "GPSLAT", "GPSLON"]].head(1)
            bot_df.columns = ["bottom_time", "latitude", "longitude"]
            bot_df.insert(0, "SSSCC", ssscc)
            add_header = not Path(fname).exists()  # add header iff file doesn't exist
            with open(fname, "a") as f:
                bot_df.to_csv(f, mode="a", header=add_header, index=False)

            mean_df.to_pickle(cfg.dirs["bottle"] + ssscc + "_btl_mean.pkl")

    return True


def convertFromSBEReader(sbeReader, ssscc):
    """Handler to convert engineering data to sci units automatically.
    Takes SBEReader object that is already connected to the .hex and .XMLCON files.
    """

    # Retrieve parsed scans and convert to dataframe
    rawData = sbeReader.parsed_scans
    raw_df = pd.DataFrame(rawData).apply(pd.to_numeric, errors="ignore")
    raw_df.index.name = "index"

    # Metadata needs to be processed seperately and then joined with the converted data
    log.info(f"Building metadata dataframe for {ssscc}")
    metaArray = [line.split(",") for line in sbeReader._parse_scans_meta().tolist()]
    meta_cols, meta_dtypes = sbeReader._breakdown_header()
    meta_df = pd.DataFrame(metaArray)
    meta_df.columns = meta_cols
    meta_df.index.name = "index"

    for col, dtype in zip(meta_cols, meta_dtypes):
        if dtype != "bool_":
            meta_df[col] = meta_df[col].astype(dtype)
        else:
            # map from string "pseudo-boolean" to actual boolean values
            meta_df[col] = meta_df[col].map({"True": True, "False": False})

    log.info("Success!")

    t_probe = meta_df["pressure_temp_int"].tolist()  # raw int from Digitquartz T probe

    # Temporary arrays to hold scientific values needed to compute cond/oxy
    t_array, p_array, c_array = [], [], []

    # needs to search sensor dictionary, and compute in order:
    # temp, pressure, cond, salinity, oxygen, all aux.
    # run one loop that builds a queue to determine order of processing, must track which column to pull
    # process queue, store results in seperate arrays for reuse later
    # once queue is empty, attach results together according to format order or xmlcon order - structure to keep track
    queue_metadata = []
    temp_counter = 0
    cond_counter = 0
    oxygen_counter = 0
    u_def_p_counter = 0
    u_def_e_counter = 0
    empty_counter = 0

    # The following are definitions for every key in the dict below:
    #
    # sensor_id = number assigned by SBE for identification in XML
    # list_id = place in XML array by SBE for determining which sensor is which, alternatively channel number (freq+volt)
    # channel_pos = is it the first, second, third, etc sensor of its type in the data file, aux sensors default to 0
    # ranking = data processing ranking - temp first, then pressure, then conductivity, then oxygen, then aux
    # column = column in the raw_df containing the engineering units to be converted to sci units
    # sensor_info = xml sensor info to convert from eng units to sci units

    rawConfig = sbeReader.parsed_config()  # Retrieve Config data

    for list_id, sensor_info in rawConfig["Sensors"].items():
        sensor_id = sensor_info["SensorID"]

        if sensor_id == "55":  # temperature block
            temp_counter += 1
            channel_pos = temp_counter
            ranking = 1

        elif sensor_id == "45":  # pressure block
            channel_pos = ""
            ranking = 2

        elif sensor_id == "3":  # conductivity block
            cond_counter += 1
            channel_pos = cond_counter
            ranking = 3

        elif sensor_id == "38":  # oxygen block
            oxygen_counter += 1
            channel_pos = oxygen_counter
            ranking = 5

        elif sensor_id == "27":  # empty block
            empty_counter += 1
            channel_pos = empty_counter
            ranking = 6

        elif sensor_id == "61":  # user defined (polynomial) block
            u_def_p_counter += 1
            channel_pos = u_def_p_counter
            ranking = 6

        elif sensor_id == "80":  # user defined (exponential) block
            u_def_e_counter += 1
            channel_pos = u_def_e_counter
            ranking = 6

        else:  # auxiliary block
            channel_pos = ""
            ranking = 7

        queue_metadata.append(
            {
                "sensor_id": sensor_id,
                "list_id": list_id,
                "channel_pos": channel_pos,
                "ranking": ranking,
                "column": list_id,
                "sensor_info": sensor_info,
            }
        )

    # Temporary block in order to append basic salinity (calculated from T1/C1) to file
    # If additional salinity is needed (e.g. T2/C2), it'll need a full reworking
    queue_metadata.append(
        {
            "sensor_id": "1000",
            "list_id": 1000,
            "channel_pos": "",
            "ranking": 4,
            "column": "",
            "sensor_info": "",
        }
    )

    # Queue sorting forces it to be in order, so we don't worry about order here
    # Assumes first channel for each sensor is primary for computing following data
    queue_metadata = sorted(queue_metadata, key=lambda sensor: sensor["ranking"])

    # Initialize converted dataframe
    converted_df = pd.DataFrame()

    for meta in queue_metadata:

        col = f"{short_lookup[meta['sensor_id']]['short_name']}{meta['channel_pos']}"
        sensor_name = short_lookup[meta["sensor_id"]]["long_name"]
        sensor_units = short_lookup[meta["sensor_id"]]["units"]
        coefs = meta["sensor_info"]

        ### Temperature block
        if meta["sensor_id"] == "55":
            log.info(f"Processing Sensor ID: {meta['sensor_id']}, {sensor_name}")
            converted_df[col] = sbe_eq.sbe3(raw_df[meta["column"]], coefs)
            if meta["list_id"] == 0:
                t_array = converted_df[col].astype(float)
                log.info(
                    f"\tPrimary temperature first reading: {t_array[0]} {sensor_units}"
                )

        ### Pressure block
        elif meta["sensor_id"] == "45":
            log.info(f"Processing Sensor ID: {meta['sensor_id']}, {sensor_name}")
            converted_df[col] = sbe_eq.sbe9(raw_df[meta["column"]], t_probe, coefs)
            if meta["list_id"] == 2:
                p_array = converted_df[col].astype(float)
                log.info(f"\tPressure first reading:  {p_array[0]} {sensor_units}")

        ### Conductivity block
        elif meta["sensor_id"] == "3":
            log.info(f"Processing Sensor ID: {meta['sensor_id']}, {sensor_name}")
            converted_df[col] = sbe_eq.sbe4(
                raw_df[meta["column"]], t_array, p_array, coefs
            )
            if meta["list_id"] == 1:
                c_array = converted_df[col].astype(float)
                log.info(f"\tPrimary cond first reading: {c_array[0]} {sensor_units}")

        ### Oxygen block
        elif meta["sensor_id"] == "38":
            log.info(f"Processing Sensor ID: {meta['sensor_id']}, {sensor_name}")
            V_corrected = sbe_eq.sbe43_hysteresis_voltage(
                raw_df[meta["column"]], p_array, coefs
            )
            converted_df[col] = sbe_eq.sbe43(
                V_corrected,
                p_array,
                t_array,
                c_array,
                coefs,
                lat=meta_df["GPSLAT"],
                lon=meta_df["GPSLON"],
            )
            converted_df["CTDOXYVOLTS"] = raw_df[meta["column"]]

        ### Fluorometer Seapoint block
        elif meta["sensor_id"] == "11":
            log.info(f"Processing Sensor ID: {meta['sensor_id']}, {sensor_name}")
            converted_df[col] = sbe_eq.seapoint_fluoro(raw_df[meta["column"]], coefs)

        ### Salinity block
        elif meta["sensor_id"] == "1000":
            log.info(f"Processing Sensor ID: {meta['sensor_id']}, {sensor_name}")
            converted_df[col] = gsw.SP_from_C(c_array, t_array, p_array)

        ### Altimeter block
        elif meta["sensor_id"] == "0":
            log.info(f"Processing Sensor ID: {meta['sensor_id']}, {sensor_name}")
            converted_df[col] = sbe_eq.sbe_altimeter(raw_df[meta["column"]], coefs)

        ### Rinko block
        elif meta["sensor_id"] == "61":
            if meta["sensor_info"]["SensorName"] in ("RinkoO2V", "RINKO", "RINKOO2", "Rinko02"):
                log.info("Processing Rinko O2")
                # hysteresis correct then pass through voltage (see Uchida, 2010)
                coefs = {"H1": 0.0065, "H2": 5000, "H3": 2000, "offset": 0}
                converted_df[col] = sbe_eq.sbe43_hysteresis_voltage(
                    raw_df[meta["column"]],
                    p_array,
                    coefs,
                )
            elif meta["sensor_info"]["SensorName"] in ("RinkoT"):
                log.info("Processing Rinko T")
                converted_df[col] = raw_df[meta["column"]]

        ### Aux block
        else:
            log.info(f"Passing along Sensor ID: {meta['sensor_id']}, {sensor_name}")
            converted_df[col] = raw_df[meta["column"]]

    # Set the column name for the index
    converted_df.index.name = "index"

    log.info("Joining metadata dataframe with converted data...")
    converted_df = converted_df.join(meta_df)
    log.info("Success!")

    # return the converted data as a dataframe
    return converted_df


def to_temperature(raw, manufacturer, sensor, coefs):
    """
    Wrapper to convert raw temperature output to scientific units using appropriate
    conversion function.

    Parameters
    ----------
    raw : array-like
        Raw temperature output
    manufacturer : str
        Name of sensor manufacturer (e.g. "seabird", "RBR")
    sensor : str
        Name of sensor (e.g. for SBE "3", "4", "9")
    coefs : dict
        Dictionary of calibration coefficient from cal sheet

    Returns
    -------
    temperature : array-like
        Converted temperature (ITS-90)

    """
    # potential example of flexible/user-friendly conversion function

    if manufacturer.lower() in ["sbe", "seabird", "sea-bird"]:
        if sensor == "3":
            # sbe_eq.sbe3()
            pass
        elif sensor == "35":
            # ?
            pass

    elif manufacturer.lower() == "rbr":
        # will need to create equations_rbr module
        if sensor.lower() == "duo":
            pass
        elif sensor.lower() == "concerto":
            pass


def CR_to_cond(cr, bath_t, ref_t, btl_p):
    """
    Convert AutoSal double conductivity ratio (CR) to conductivity using
    GSW conversion routines.

    Parameters
    ----------
    cr : array-like
        Double conductivity ratio from AutoSal, unitless
    bath_t : array-like
        AutoSal water bath temperature, degrees C
    ref_t : array-like
        CTD temperature at bottle stop, degrees C
    btl_p : array-like
        CTD pressure at bottle stop, dbar

    Returns
    -------
    cond : array-like
        Converted reference conductivity, mS/cm

    """

    salinity = gsw.SP_salinometer((cr / 2.0), bath_t)
    cond = gsw.C_from_SP(salinity, ref_t, btl_p)

    # ignore RunTimeWarning from (np.nan <= 1)
    with np.errstate(invalid="ignore"):
        cond[cond <= 1] = np.nan

    return cond



import logging
from pathlib import Path, PosixPath, WindowsPath

from click.testing import CliRunner

import ctdcal.__main__ as main


def test_cli():
    runner = CliRunner()
    result = runner.invoke(main.cli)
    assert "cruise-report" in result.output
    assert "import" in result.output
    assert "init" in result.output
    assert "process" in result.output
    assert "qc" in result.output


def test_debug(tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        # default options
        result = runner.invoke(main.cli, "process")
        assert "Debug mode off" in result.output

        # no debug
        result_no_debug = runner.invoke(main.cli, ["--no-debug", "process"])
        assert "Debug mode off" in result_no_debug.output

        # debug (run last so Logger("ctdcal").level=NOTSET for other tests)
        result_debug = runner.invoke(main.cli, ["--debug", "process"])
        assert "Debug mode on" in result_debug.output


def test_init(tmp_path, caplog):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        # check first init call creates dirs successfully
        with caplog.at_level(logging.INFO):
            result = runner.invoke(main.init)
        assert "Building default /data/ directories" in caplog.messages[0]
        assert result.exit_code == 0
        assert not result.exception

        # check "folders already exist" error
        caplog.clear()
        with caplog.at_level(logging.INFO):
            result_rerun = runner.invoke(main.init)
        assert "Building default /data/ directories" in caplog.messages[0]
        assert result_rerun.exit_code == 1
        assert isinstance(result_rerun.exception, FileExistsError)
        if isinstance(Path.cwd(), PosixPath):
            assert "data/ssscc" in result_rerun.exception.filename
        elif isinstance(Path.cwd(), WindowsPath):
            assert "data\\ssscc" in result_rerun.exception.filename


def test_import_data():
    runner = CliRunner()
    result = runner.invoke(main.import_data)
    assert result.exit_code == 1
    assert isinstance(result.exception, NotImplementedError)


def test_process(tmp_path, caplog):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        # default options
        with caplog.at_level(logging.INFO):
            result = runner.invoke(main.process)
        assert result.exit_code == 1
        # assert isinstance(result.exception, FileNotFoundError)
        # assert "ssscc.csv" in result.exception.filename
        # assert "generating from .hex file list" in caplog.messages[-1]

        # PMEL option
        with caplog.at_level(logging.INFO):
            result_PMEL = runner.invoke(main.process, ["-g", "PMEL"])
        assert result_PMEL.exit_code == 1
        assert isinstance(result_PMEL.exception, NotImplementedError)


def test_cruise_report(tmp_path, caplog):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        with caplog.at_level(logging.INFO):
            result = runner.invoke(main.cruise_report)
        assert isinstance(result.exception, FileNotFoundError)
        assert "report_data.csv" in result.exception.filename


def test_quick_convert(tmp_path, caplog):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        with caplog.at_level(logging.INFO):
            result = runner.invoke(main.quick_convert)
        assert result.exit_code == 0
        assert "Found 0 .cnv files" in caplog.messages[0]

        caplog.clear()
        with caplog.at_level(logging.INFO):
            result_ctd = runner.invoke(main.quick_convert, ["-f", "ct1"])
        assert result_ctd.exit_code == 0
        assert "Found 0 .cnv files" in caplog.messages[0]

        caplog.clear()
        with caplog.at_level(logging.INFO):
            result_btl = runner.invoke(main.quick_convert, ["-f", "hy1"])
        assert result_btl.exit_code == 1
        assert isinstance(result_btl.exception, NotImplementedError)



"""
A module for handling ODF input-output files related to CTDCAL.
"""

import logging
from io import BufferedIOBase, BytesIO, StringIO
from pathlib import Path
from typing import Union
from zipfile import ZipFile, is_zipfile
from zipimport import ZipImportError

import pandas as pd
import requests

log = logging.getLogger(__name__)


def load_cnv(cnv_file: Union[str, Path]) -> pd.DataFrame:
    """
    Load Sea-Bird converted (.cnv) cast file into DataFrame
    """
    with open(cnv_file) as f:
        file = f.readlines()

    # parse column names
    info = dict()
    cols = []
    for idx, line in enumerate(file):
        # get variable info
        if line.strip("# \n").startswith(("nquan", "nvalues", "units", "bad_flag")):
            k, v = line.strip("# \n").split("=")
            info[k.strip()] = v.strip()

        # get column names
        elif line.strip("# ").startswith("name"):
            # expected format is:   # name 0 = col_name: long_description
            cols.append(line.split(":")[0].split("=")[-1].strip())

        # last row before data begins
        elif line.startswith("*END*"):
            data_index = idx + 1
            break

        # anything else is a comment line
        else:
            continue  # pragma: no cover

    # read data
    return pd.read_csv(
        cnv_file,
        skiprows=range(0, data_index),
        # delim_whitespace=True,
        sep=r'\s+',
        names=cols,
        engine="python",
        skipinitialspace=True,
        na_values=info["bad_flag"],
    )


def load_exchange_btl(btl_file: Union[str, Path]) -> pd.DataFrame:
    """
    Load WHP-exchange bottle file (_hy1.csv) into DataFrame. File can be on local
    file system or downloaded from an appropriate cchdo.ucsd.edu link
    (e.g., https://cchdo.ucsd.edu/data/19436/325020210316_hy1.csv)

    Adapted from cchdo.hydro package.

    Parameters
    ----------
    btl_file : str or Path
        Name or URL of file to be loaded

    Returns
    -------
    df : DataFrame
        Loaded bottle file
    """
    # read from url
    if isinstance(btl_file, (str, Path)) and str(btl_file).startswith("http"):
        log.info(f"Loading bottle file {Path(btl_file).name} from http link")
        file = requests.get(btl_file).text.splitlines(keepends=True)

    # read from file
    elif isinstance(btl_file, (str, Path)):
        log.info(f"Loading bottle file {Path(btl_file).name} from local file")
        with open(btl_file) as f:
            file = f.readlines()

    # find index of units row
    for idx, line in enumerate(file):
        # skip comment lines (which may reference EXPOCODE and break membership test)
        if line.strip().startswith("#"):
            continue

        # find index of units row
        if "EXPOCODE" in line:
            units = idx + 1  # units row immediately follows column names
            break

    return pd.read_csv(
        StringIO("".join(file)),
        skiprows=[0, units],
        skipfooter=1,
        engine="python",
        comment="#",
        skipinitialspace=True,
    )


def load_exchange_ctd(
    ctd_file: Union[str, Path, BufferedIOBase],
    n_files=None,
    recursed=False,
) -> pd.DataFrame:
    """
    Load WHP-exchange CTD file(s) (_ct1.csv) into DataFrame. File(s) can be on local
    file system or downloaded from an appropriate cchdo.ucsd.edu link
    (e.g., https://cchdo.ucsd.edu/data/19434/325020210316_ct1.zip)

    Adapted from cchdo.hydro package.

    Parameters
    ----------
    ctd_file : str or Path
        Name or URL of file to be loaded

    n_files : int, optional
        Number of files to load from .zip archive

    Returns
    -------
    header : dict or list of dict
        File metadata from header(s) (e.g., EXPOCODE, STNNBR, CASTNO)
    df : DataFrame or list of DataFrame
        Loaded CTD file(s)
    """
    # read from url (.zip)
    if isinstance(ctd_file, (str, Path)) and str(ctd_file).startswith("http"):
        log.info(f"Loading CTD file {Path(ctd_file).name} from http link")
        data_raw = BytesIO(requests.get(ctd_file).content)

    # read from file
    elif isinstance(ctd_file, (str, Path)):
        log.info(f"Loading CTD file {Path(ctd_file).name} from local file")
        with open(ctd_file, "rb") as f:
            data_raw = BytesIO(f.read())

    # read from open file
    elif isinstance(ctd_file, BufferedIOBase):
        log.info("Loading open file object")
        data_raw = BytesIO(ctd_file.read())

    # .zip special behavior
    if is_zipfile(data_raw):
        log.info("Loading CTD files from .zip")

        if recursed is True:
            raise ZipImportError("Recursive .zip files encountered... exiting")

        data_raw.seek(0)  # is_zipfile moves cursor to EOF, reset to start
        zip_contents = []
        with ZipFile(data_raw) as zf:
            for zipinfo in zf.infolist():
                zip_contents.append(BytesIO(zf.read(zipinfo)))

        # list comprehension is same as using functools.partial, just different syntax
        return zip(
            *[load_exchange_ctd(zc, recursed=True) for zc in zip_contents[:n_files]]
        )

    else:
        data_raw.seek(0)  # is_zipfile moves cursor to EOF, reset to start
        file = data_raw.read().decode("utf8").splitlines(keepends=True)

    # process metadata
    for idx, line in enumerate(file):
        # skip comment lines (which may reference CTDPRS and break membership test)
        if line.strip().startswith("#"):
            continue

        # find header info
        if line.startswith("NUMBER_HEADERS"):
            header_ind = idx

        # find index of units row
        if "CTDPRS" in line:
            columns = idx
            units = idx + 1  # units row immediately follows column names
            break

    # break down header rows
    header = {}
    for line in file[header_ind:columns]:
        k, v = line.strip("\n").split("=")
        header[k.strip()] = v.strip()

    return header, pd.read_csv(
        StringIO("".join(file)),
        skiprows=list(range(0, columns)) + [units],  # skip up to column names (+ units)
        skipfooter=1,
        engine="python",
        comment="#",
        skipinitialspace=True,
    )



"""
A module for plotting and visualizing the CTDCAL data products.
"""
import logging
from pathlib import Path

import gsw
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import pandas as pd

log = logging.getLogger(__name__)


def _apply_default_fmt(xlim, ylim, xlabel, ylabel, title, grid, fontsize=12):
    plt.xlim(xlim)
    plt.xticks(rotation=45)
    plt.ylim(ylim)
    plt.xlabel(xlabel, fontsize=fontsize)
    plt.ylabel(ylabel, fontsize=fontsize)
    plt.title(title, fontsize=fontsize)
    plt.grid(grid)
    plt.tight_layout()


def _save_fig(ax, f_out):
    if f_out is not None:
        f_parent = Path(f_out).parent
        if not f_parent.exists():
            log.info(f"Parent folder '{f_parent.as_posix()}' doesn't exist... creating")
            Path(f_out).parent.mkdir(parents=True)
        plt.savefig(f_out)
        plt.close()
    else:
        return ax


def residual_vs_pressure(
    param,
    ref,
    prs,
    stn=None,
    xlim=(-0.02, 0.02),
    ylim=(6000, 0),
    xlabel="Residual",
    ylabel="Pressure (dbar)",
    auto_title=True,
    grid=False,
    deep=False,
    f_out=None,
):
    """
    Plot residuals (ref - param) as a function of pressure.

    Parameters
    ----------
    param : pd.Series or array-like
        Input variable
    ref : pd.Series or array-like
        Reference variable
    prs : pd.Series or array-like
        Pressure variable
    stn : pd.Series or array-like, optional
        Station variable
    xlim : tuple, optional
        Lower and upper x-limits
    ylim : tuple, optional
        Lower and upper y-limits
    xlabel : str, optional
        Label for the x-axis
    ylabel : str, optional
        Label for the y-axis
    auto_title : bool, optional
        Generate title from input attributes (iff dtype is pd.Series)
    grid : bool, optional
        Overlay grid on figure
    deep : bool, optional
        Whether to plot only values >2000 dbar
    f_out : path-like, optional
        Path to save figure

    Returns
    -------
    ax : matplotlib.axes, optional
        Formatted scatter plot (if f_out is not given)
    """
    diff = ref - param
    deep_text = " (>2000 dbar)" if deep else ""
    if deep:
        deep_rows = prs > 2000
        diff, prs, stn = diff[deep_rows], prs[deep_rows], stn[deep_rows]

    # initialize figure
    plt.figure(figsize=(7, 6))
    ax = plt.axes()

    # color scatter by stations if given
    if stn is not None:
        idx, uniques = pd.factorize(stn)  # find unique stations #s and index them
        sc = ax.scatter(diff, prs, c=idx, marker="+")
        cbar = plt.colorbar(sc, ax=ax, pad=0.1)  # set cbar ticks to station names
        tick_inds = cbar.get_ticks().astype(int)
        cbar.ax.yaxis.set_major_locator(ticker.FixedLocator(tick_inds))
        cbar.ax.set_yticklabels(uniques[tick_inds])
        cbar.ax.set_title("Station")
    else:
        sc = ax.scatter(diff, prs, marker="+")

    # formatting
    title = None
    if auto_title:
        try:
            title = f"{ref.name}-{param.name}{deep_text} vs. {prs.name}"
        except AttributeError:
            log.warning(
                "Failed to set title from variable names (requires dtype pd.Series)"
            )
            log.info("Set afterward using 'ax.set_title(\"title\")'")
    _apply_default_fmt(xlim, ylim, xlabel, ylabel, title, grid)

    # save to path or return axis
    return _save_fig(ax, f_out)


def residual_vs_station(
    param,
    ref,
    prs,
    stn,
    ylim=(-0.02, 0.02),
    xlabel="Station Number",
    ylabel="Residual",
    grid=False,
    deep=False,
    f_out=None,
):
    """
    Plot residuals (ref - param) as a function of station number.

    Parameters
    ----------
    param : pd.Series or array-like
        Input variable
    ref : pd.Series or array-like
        Reference variable
    prs : pd.Series or array-like
        Pressure variable
    stn : pd.Series or array-like, optional
        Station variable
    ylim : tuple, optional
        Lower and upper y-limits
    xlabel : str, optional
        Label for the x-axis
    ylabel : str, optional
        Label for the y-axis
    grid : bool, optional
        Overlay grid on figure
    deep : bool, optional
        Whether to plot only values >2000 dbar
    f_out : path-like, optional
        Path to save figure

    Returns
    -------
    ax : matplotlib.axes, optional
        Formatted scatter plot (if f_out is not given)
    """
    diff = ref - param
    deep_text = " (>2000 dbar)" if deep else ""
    if deep:
        deep_rows = prs > 2000
        diff, prs, stn = diff[deep_rows], prs[deep_rows], stn[deep_rows]

    plt.figure(figsize=(7, 6))
    ax = plt.axes()
    sc = ax.scatter(stn, diff, c=prs, marker="+")
    cbar = plt.colorbar(sc, pad=0.1)
    cbar.set_label("Pressure (dbar)")

    # formatting
    try:
        title = f"{ref.name}-{param.name}{deep_text} vs. {stn.name}"
    except AttributeError:
        title = None
        log.warning(
            "Failed to set title from variable names (requires dtype pd.Series)"
        )
        log.info('Set afterward using \'ax.set_title("title")`')
    _apply_default_fmt(None, ylim, xlabel, ylabel, title, grid)

    # save to path or return axis
    return _save_fig(ax, f_out)


def _intermediate_residual_plot(
    diff,
    prs,
    stn,
    xlim=(-0.02, 0.02),
    xlabel="Residual",
    show_thresh=False,
    f_out=None,
):
    """
    Internal function to make figures at intermediate processing stages for debugging.
    """
    ax = residual_vs_pressure(
        0, diff, prs, stn=stn, xlim=xlim, xlabel=xlabel, auto_title=False, grid=True
    )

    if show_thresh:
        thresh = np.array([0.002, 0.005, 0.010, 0.020])
        p_range = np.array([6000, 2000, 1000, 500])
        thresh = np.append(thresh, thresh[-1])  # this should still work fine even when
        p_range = np.append(p_range, 0)  # thresh/p_range are defined elsewhere
        plt.step(thresh, p_range, ":k")
        plt.step(-thresh, p_range, ":k")

    mean = np.round(np.nanmean(diff), 4)
    stdev = np.round(np.nanstd(diff), 4)
    ax.set_title(f"Mean: {mean} / Stdev: {stdev}")

    # save to path or return axis (primarily for testing)
    return _save_fig(ax, f_out)


def param_vs_param(param1, label1, param2, label2, f_out=None, stn=None, tsT=None, tsS=None):
    """
    Scatter plots two CTD parameters against each other with support for TS plotting.

    Support for bottle or time data.

    Most TS plots have temperature on the Y axis and salinity on the X axis.
    
    Parameters
    ----------
    param1 : array-like of float
        The first column/array for plotting on the X axis
    label1 : string
        Label for the X axis, attributed to param1
    param2 : array-like of float
        The second column/array for plotting on the Y-axis
    label2 : string
        Label for the Y axis, attributed to param2
    f_out : String, optional
        Path and filename to save TS plot
    stn : array-like, optional
        An array of station numbers or labels for 
    tsT : array-like of float, optional
        Temperature array for use in density gradient calculations
    tsS : array-like of float, optional
        Salinity array for use in density gradient calculations
    """

    fig = plt.figure(figsize=(7, 6))
    ax = fig.add_subplot()

    if stn is not None:
        stn = np.array(stn)  # Ensure stn is a NumPy array
        sc = ax.scatter(param1, param2, c=stn.astype(int), marker="+")
        sc.set_clim(vmin=stn.astype(int).min(), vmax=stn.astype(int).max()) #   Color mapping based on station number
        cbar = plt.colorbar(sc, format=ticker.FormatStrFormatter("%.0f"))
        cbar.set_label("Station/Cast Label")
    else:
        sc = ax.scatter(param1, param2, color='black', marker="+")

    ax.set_xlabel(label1)
    ax.set_ylabel(label2)

    #   TS plotting - define density contours
    if tsT is not None:
        #   Demarcate temperature, salinty, and limits for density
        smin = tsS.min() - (0.01 * tsS.min())
        smax = tsS.max() + (0.01 * tsS.max())
        tmin = tsT.min() - (0.1 * tsT.max())
        tmax = tsT.max() + (0.1 * tsT.max())
        #   Gridding for a contour
        xdim = round((smax - smin) / 0.1 + 1.0)
        ydim = round((tmax - tmin) + 1.0)
        #   Creating density vectors to fill the grid with densities
        dens = np.zeros((ydim, xdim))
        si = np.linspace(1, xdim - 1, xdim) * 0.1 + smin
        ti = np.linspace(1, ydim - 1, ydim) + tmin
        for j in range(0, int(ydim)):
            for i in range(0, int(xdim)):
                dens[j, i] = gsw.rho(si[i], ti[j], 0)
        dens = dens - 1000  #   Convert to sigma-theta
        #   Density contours formatting
        cs = plt.contour(si, ti, dens, linestyles="dashed", colors="k")
        plt.clabel(cs, fontsize=12, inline=1, fmt="%0.1f")

    return _save_fig(ax, f_out) #   Save the figure if f_out is given else return the scatter axis



"""
A module for handling Autosalinomter files, in the ODF format (Carl Mattson).
"""

import csv
import io
import logging
from pathlib import Path

import gsw
import numpy as np
import pandas as pd

from ctdcal import get_ctdcal_config
from ctdcal.common import validate_file
from ctdcal.fitting.common import (
    NodeNotFoundError,
    df_node_to_BottleFlags,
    get_node,
    save_node,
)

cfg = get_ctdcal_config()
log = logging.getLogger(__name__)


def _salt_loader(filename):
    """
    Load raw ODF salt file into salt, reference, and flagged DataFrames.

    ODF salt files have a header of configuration information, including the K15 and run date.
    """
    csv_opts = dict(delimiter=" ", quoting=csv.QUOTE_NONE, skipinitialspace="True")
    if isinstance(filename, (str, Path)):
        with open(filename, newline="") as f:
            saltF = csv.reader(f, **csv_opts)
            saltArray = [row for row in saltF]
            ssscc = Path(filename).stem
    elif isinstance(filename, io.StringIO):
        saltF = csv.reader(filename, **csv_opts)
        saltArray = [row for row in saltF]
        ssscc = "test_odf_io"
    else:
        raise NotImplementedError(
            "Salt loader only able to read in str, Path, or StringIO classes"
        )

    #   Check on the header
    date = pd.to_datetime(saltArray[0][-1])
    k15  = float(saltArray[0][8])
    if ((date < pd.Timestamp('1985-01-01')) | (date > pd.Timestamp.today())):
        #   Check on the run date - should this be ± a few weeks to confirm that the analyst's computer is set up right?
        log.warning(f"Salt file for {ssscc} has an erroneous timestamp in the header.")
    if ((int(saltArray[0][-2]) < 0) | int(saltArray[0][-2]) > 1000):
        #   Standardization dial ranges from 0 - 1000
        log.warning(f"Salt file for {ssscc} has an erroneous standardization dial in the header.")

    del saltArray[0]  # remove file header
    saltDF = pd.DataFrame.from_records(saltArray)

    cols = dict(  # having this as a dict streamlines next steps
        [
            ("STNNBR", int),
            ("CASTNO", int),
            ("SAMPNO", int),
            ("BathTEMP", int),
            ("CRavg", float),
            ("autosalSAMPNO", int),
            ("Unknown", int),
            ("StartTime", object),
            ("EndTime", object),
            ("Attempts", int),
        ]
    )

    # add as many "Reading#"s as needed
    for ii in range(0, len(saltDF.columns) - len(cols)):
        cols["Reading{}".format(ii + 1)] = float
    if "Reading6" in cols:
        log.warning("More than 5 readings found in salts file - check columns of source.")
    saltDF.columns = list(cols.keys())  # name columns

    # check for commented out lines
    commented = saltDF["STNNBR"].str.startswith(("#", "x"))
    if commented.any():
        log.debug(f"Found comment character (#, x) in {ssscc} salt file, ignoring line")
        saltDF = saltDF[~commented]
    if saltDF["autosalSAMPNO"].value_counts().get('worm',0) != 2:
        #   Confirm 2-point standardization
        log.warning(f"Suspect number of standards found in salts file {ssscc}. Confirm standardizations.")    
    checkvals = saltDF[saltDF.autosalSAMPNO == "worm"].CRavg.astype(float)
    if ((checkvals >= 1.0001 * 2 * k15) & (checkvals <= 0.9999 * 2 * k15)).any():
        #   Check the worm values rel. to batch K15 - within 1/100 %
        log.warning(f"Standards in raw salt file {ssscc} are >0.01 % off from expected K15 value.")
    if any((saltDF['BathTEMP'].astype(float) < 17.5) | (saltDF['BathTEMP'].astype(float) > 33.5)):
        #   Check bath temperature
        log.warning(f"Raw salt file for {ssscc} contains bath temperatures outside of operational ranges.")

    # check end time for * and code questionable
    # (unconfirmed but * appears to indicate a lot of things from LabView code:
    # large spread in values, long time between samples, manual override, etc.)
    flagged = saltDF["EndTime"].str.contains("*", regex=False)
    questionable = None
    if flagged.any():
        # remove asterisks from EndTime and flag samples
        log.debug(f"Found * in {ssscc} salt file, flagging value(s) as questionable")
        saltDF["EndTime"] = saltDF["EndTime"].str.strip("*")
        # construct flagged dataframe
        questionable = pd.DataFrame()
        questionable["bottle_num"] = saltDF.loc[flagged, "SAMPNO"].astype(int)
        questionable["cast_id"] = [str(s).zfill(3)[-3:] + str(c).zfill(2) for s, c in zip(saltDF.loc[flagged, "STNNBR"], saltDF.loc[flagged, "CASTNO"])]
        questionable["value"] = 3
        questionable["notes"] = "Auto-flagged by processing function (had * in row)"

    # add time (in seconds) needed for autosal drift removal step
    saltDF["IndexTime"] = pd.to_datetime(saltDF["EndTime"], format="%H:%M:%S")
    saltDF["IndexTime"] = (saltDF["IndexTime"] - saltDF["IndexTime"].iloc[0]).dt.seconds
    saltDF["IndexTime"] += (saltDF["IndexTime"] < 0) * (3600 * 24)  # fix overnight runs

    refDF = saltDF.loc[
        saltDF["autosalSAMPNO"] == "worm", ["IndexTime", "CRavg"]
    ].astype(float)
    saltDF = saltDF[saltDF["autosalSAMPNO"] != "worm"].astype(cols)  # force dtypes

    return saltDF, refDF, questionable


def remove_autosal_drift(saltDF, refDF):
    """Calculate linear CR drift between reference values"""
    if refDF.shape != (2, 2):
        ssscc = f"{saltDF['STNNBR'].unique()[0]:03d}{saltDF['CASTNO'].unique()[0]:02d}"
        log.warning(
            f"Failed to find start/end reference readings for {ssscc}, check salt file"
        )
    else:
        # find rate of drift
        diff = refDF.diff(axis="index").dropna()

        if any(abs(diff["CRavg"] > 0.0001)):
            #   Warn analyst that the standard drifted by a lot (precise to within 5 decimal places)
            log.warning(f"Salt run at station {saltDF['STNNBR'].iloc[0]} had a CR drift in excess of 0.0001.")
        if any(diff["IndexTime"] > 43200):
            #   Warn analyst that over 12 hours passed between standardizations
            log.warning(f"Salt run at station {saltDF['STNNBR'].iloc[0]} had >12 hours between standardizations.")
        if any(saltDF["IndexTime"].diff().sum() > (diff["IndexTime"])):
            #   Warn analyst that standardization and sample timestamps may not be aligned
            #   Standards should always bookend the samples
            log.warning(f"Salt run at station {saltDF['STNNBR'].iloc[0]} has misaligned standards and samples.")

        time_coef = (diff["CRavg"] / diff["IndexTime"]).iloc[0]

        # apply offset as a linear function of time
        saltDF = saltDF.copy(deep=True)  # avoid modifying input dataframe
        saltDF["CRavg"] += saltDF["IndexTime"] * time_coef
        saltDF["CRavg"] = saltDF["CRavg"].round(5)  # match initial precision

    return saltDF.drop(labels="IndexTime", axis="columns")


def _salt_exporter(
    saltDF, outdir=cfg.dirs["salt"], stn_col="STNNBR", cast_col="CASTNO"
):
    """
    Export salt DataFrame to .csv file. Extra logic is included in the event that
    multiple stations and/or casts are included in a single raw salt file.
    """
    stations = saltDF[stn_col].unique()
    if len(stations) > 1:
        log.info("Multiple stations found in salt source file. Writing out .CSVs seperately.")

    for station in stations:
        stn_salts = saltDF[saltDF[stn_col] == station]
        casts = stn_salts[cast_col].unique()
        for cast in casts:
            stn_cast_salts = stn_salts[stn_salts[cast_col] == cast].copy()
            stn_cast_salts.dropna(axis=1, how="all", inplace=True)  # drop empty columns
            outfile = Path(outdir) / f"{station:03.0f}{cast:02.0f}_salts.csv"  # SSSCC_*
            if outfile.exists():
                log.info(str(outfile) + " already exists...skipping")
                continue
            stn_cast_salts.to_csv(outfile, index=False)


def process_salts(ssscc_list, user_cfg=None, salt_dir=cfg.dirs["salt"]):
    """
    Master salt processing function. Load in salt files for given station/cast list,
    calculate salinity, and export to .csv files.

    Parameters
    ----------
    ssscc_list : list of str
        List of stations to process
    user_cfg : Munch object
        Dictionary of user configuration parameters
    salt_dir : str, optional
        Path to folder containing raw salt files (defaults to data/salt/)

    """
    flags_df = None
    for ssscc in ssscc_list:
        if (Path(salt_dir) / f"{ssscc}_salts.csv").exists():
            log.info(f"{ssscc}_salts.csv already exists in {salt_dir}... skipping")
            continue
        else:
            try:
                saltDF, refDF, questionable = _salt_loader(Path(salt_dir) / ssscc)
            except FileNotFoundError:
                log.warning(f"Salt file for cast {ssscc} does not exist... skipping")
                continue
            saltDF = remove_autosal_drift(saltDF, refDF)
            saltDF["SALNTY"] = gsw.SP_salinometer(
                (saltDF["CRavg"] / 2.0), saltDF["BathTEMP"]
            )  # .round(4)
            _salt_exporter(saltDF, salt_dir)

            # compile flags
            if questionable is not None:
                if flags_df is None:
                    flags_df = questionable
                else:
                    flags_df = pd.concat([flags_df, questionable], ignore_index=True)



    # save flags
    if flags_df is not None:
        if user_cfg is not None:
            flag_path = Path(user_cfg.datadir, 'flag', user_cfg.bottleflags_man)
        else:   #   No user_cfg, use get_ctdcal_config TODO: Remember to remove all the old cfg calls
            flag_path = Path(cfg.dirs('flags'), "bottleflags_man.csv")
        flag_file = validate_file(flag_path, create=True)
        try:
            salt = pd.DataFrame.from_dict(get_node(flag_file, 'salt'))
        except NodeNotFoundError:
            salt = None
        flags_df = pd.concat([flags_df, salt], ignore_index=True)
        new_flags = df_node_to_BottleFlags(flags_df)
        save_node(flag_file, new_flags, 'salt', create_new=True)


def print_progress_bar(
    iteration,
    total,
    prefix="",
    suffix="",
    decimals=1,
    length=100,
    fill="█",
    printEnd="\r",
):
    """
    A progress bar, helpful for implementing into loops or highlighting progression through processing.
    
    https://stackoverflow.com/questions/3173320/text-progress-bar-in-terminal-with-block-characters/13685020
    credit: u/Greenstick
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end=printEnd)  #   Potential to add to log
    # Print New Line on Complete
    if iteration == total:
        print()



"""
A module for SBE conversion equations and related helper equations.

SBE internally references each sensor with an assigned SensorID for indexing. The
given sensor numbers are determined empirically from .XMLCON files in 2016-2017 and
not via an official document, and may change according to SBE wishes.
"""

import inspect
import logging

import gsw
import numpy as np

from ctdcal.oxy_fitting import oxy_umolkg_to_ml

log = logging.getLogger(__name__)


def _check_coefs(coefs_in, expected):
    """Compare function input coefs with expected"""
    missing_coefs = sorted(set(expected) - set(coefs_in))
    if missing_coefs != []:
        raise KeyError(f"Coefficient dictionary missing keys: {missing_coefs}")


def _check_freq(freq):
    """Convert to np.array, NaN out zeroes, convert to float if needed"""
    freq = np.array(freq)
    sensor = inspect.stack()[1].function  # name of function calling _check_freq

    if freq.dtype != float:  # can sometimes come in as object
        log.warning(f"Attempting to convert {freq.dtype} to float for {sensor}")
        freq = freq.astype(float)

    if 0 in freq:
        N_zeroes = (freq == 0).sum()
        log.warning(
            f"Found {N_zeroes} zero frequency readings in {sensor}, replacing with NaN"
        )
        freq[freq == 0] = np.nan

    return freq


def _check_volts(volts, v_min=0, v_max=5):
    """Convert to np.array, NaN out values outside of 0-5V, convert to float if needed"""
    volts = np.array(volts)
    sensor = inspect.stack()[1].function  # name of function calling _check_volts

    if volts.dtype != float:  # can sometimes come in as object
        log.warning(f"Attempting to convert {volts.dtype} to float for {sensor}")
        volts = volts.astype(float)

    if any(volts < v_min) or any(volts > v_max):
        log.warning(
            f"{sensor} has values outside of {v_min}-{v_max}V, replacing with NaN"
        )
        volts[volts < v_min] = np.nan
        volts[volts > v_max] = np.nan

    return volts


def sbe3(freq, coefs, decimals=4):
    """
    SBE equation for converting SBE3 frequency to temperature.
    SensorID: 55

    Parameters
    ----------
    freq : array-like
        Raw frequency (Hz)
    coefs : dict
        Dictionary of calibration coefficients (G, H, I, J, F0)

    Returns
    -------
    t_ITS90 : array-like
        Converted temperature (ITS-90)
    """
    _check_coefs(coefs, ["G", "H", "I", "J", "F0"])
    freq = _check_freq(freq)

    t_ITS90 = (
        1
        / (
            coefs["G"]
            + coefs["H"] * (np.log(coefs["F0"] / freq))
            + coefs["I"] * np.power((np.log(coefs["F0"] / freq)), 2)
            + coefs["J"] * np.power((np.log(coefs["F0"] / freq)), 3)
        )
        - 273.15
    )
    return np.around(t_ITS90, decimals)


def sbe4(freq, t, p, coefs, decimals=4):
    """
    SBE equation for converting SBE4 frequency to conductivity. This conversion
    is valid for both SBE4C (profiling) and SBE4M (mooring).
    SensorID: 3

    Parameters
    ----------
    freq : array-like
        Raw frequency (Hz)
    t : array-like
        Converted temperature (ITS-90 degrees C)
    p : array-like
        Converted pressure (dbar)
    coefs : dict
        Dictionary of calibration coefficients (G, H, I, J, CPcor, CTcor)

    Returns
    -------
    c_mS_cm : array-like
        Converted conductivity (mS/cm)
    """
    _check_coefs(coefs, ["G", "H", "I", "J", "CPcor", "CTcor"])
    freq_kHz = _check_freq(freq) * 1e-3  # equation expects kHz

    c_S_m = (
        coefs["G"]
        + coefs["H"] * np.power(freq_kHz, 2)
        + coefs["I"] * np.power(freq_kHz, 3)
        + coefs["J"] * np.power(freq_kHz, 4)
    ) / (10 * (1 + coefs["CTcor"] * np.array(t) + coefs["CPcor"] * np.array(p)))
    c_mS_cm = c_S_m * 10  # S/m to mS/cm

    return np.around(c_mS_cm, decimals)


def sbe9(freq, t_probe, coefs, decimals=4):
    """
    SBE/STS(?) equation for converting SBE9 frequency to pressure.
    SensorID: 45

    Parameters
    ----------
    freq : array-like
        Raw frequency (Hz)
    t_probe : array-like
        Raw integer measurement from the Digiquartz temperature probe
    coefs : dict
        Dictionary of calibration coefficients
        (T1, T2, T3, T4, T5, C1, C2, C3, D1, D2, AD590M, AD590B)

    Returns
    -------
    p_dbar : array-like
        Converted pressure (dbar)
    """
    _check_coefs(
        coefs,
        (
            ["T1", "T2", "T3", "T4", "T5"]
            + ["C1", "C2", "C3"]
            + ["D1", "D2"]
            + ["AD590M", "AD590B"]
        ),
    )
    freq_MHz = _check_freq(freq) * 1e-6  # equation expects MHz
    t_probe = (coefs["AD590M"] * np.array(t_probe).astype(int)) + coefs["AD590B"]

    T0 = (
        coefs["T1"]
        + coefs["T2"] * t_probe
        + coefs["T3"] * np.power(t_probe, 2)
        + coefs["T4"] * np.power(t_probe, 3)
    )
    w = 1 - T0 * T0 * freq_MHz * freq_MHz
    p_dbar = 0.6894759 * (
        (coefs["C1"] + coefs["C2"] * t_probe + coefs["C3"] * t_probe * t_probe)
        * w
        * (1 - (coefs["D1"] + coefs["D2"] * t_probe) * w)
        - 14.7
    )
    return np.around(p_dbar, decimals)


def sbe_altimeter(volts, coefs, decimals=1):
    """
    SBE equation for converting altimeter voltages to meters. This conversion
    is valid for altimeters integrated with any Sea-Bird CTD (e.g. 9+, 19, 25).
    Sensor ID: 0

    Parameters
    ----------
    volts : array-like
        Raw voltages
    coefs : dict
        Dictionary of calibration coefficients (ScaleFactor, Offset)

    Returns
    -------
    bottom_distance : array-like
        Distance from the altimeter to an object below it (meters)

    Notes
    -----
    Equation provdided by SBE in Application Note 95, page 1.

    While the SBE documentation refers to a Teledyne Benthos or Valeport altimeter,
    the equation works for all altimeters typically found in the wild.
    """
    _check_coefs(coefs, ["ScaleFactor", "Offset"])
    volts = _check_volts(volts)

    bottom_distance = (300 * volts / coefs["ScaleFactor"]) + coefs["Offset"]

    return np.around(bottom_distance, decimals)


def sbe43(volts, p, t, c, coefs, lat=0.0, lon=0.0, decimals=4):
    # NOTE: lat/lon = 0 is not "acceptable" for GSW, come up with something else?
    """
    SBE equation for converting SBE43 engineering units to oxygen (ml/l).
    SensorID: 38

    Parameters
    ----------
    volts : array-like
        Raw voltage
    p : array-like
        Converted pressure (dbar)
    t : array-like
        Converted temperature (Celsius)
    c : array-like
        Converted conductivity (mS/cm)
    coefs : dict
        Dictionary of calibration coefficients (Soc, offset, Tau20, A, B, C, E)
    lat : array-like, optional
        Latitude (decimal degrees north)
    lon : array-like, optional
        Longitude (decimal degrees)

    Returns
    -------
    oxy_ml_l : array-like
        Converted oxygen (mL/L)
    """
    _check_coefs(coefs, ["Soc", "offset", "Tau20", "A", "B", "C", "E"])
    volts = _check_volts(volts)
    t_Kelvin = np.array(t) + 273.15

    SP = gsw.SP_from_C(c, t, p)
    SA = gsw.SA_from_SP(SP, p, lon, lat)
    CT = gsw.CT_from_t(SA, t, p)
    sigma0 = gsw.sigma0(SA, CT)
    o2sol = gsw.O2sol(SA, CT, p, lon, lat)  # umol/kg
    o2sol_ml_l = oxy_umolkg_to_ml(o2sol, sigma0)  # equation expects mL/L

    # NOTE: lat/lon always required to get o2sol (and need SA/CT for sigma0 anyway)
    # the above is equivalent to:
    # pt = gsw.pt0_from_t(SA, t, p)
    # o2sol = gsw.O2sol_SP_pt(s, pt)

    oxy_ml_l = (
        coefs["Soc"]
        * (volts + coefs["offset"])
        * (
            1.0
            + coefs["A"] * np.array(t)
            + coefs["B"] * np.power(t, 2)
            + coefs["C"] * np.power(t, 3)
        )
        * o2sol_ml_l
        * np.exp(coefs["E"] * np.array(p) / t_Kelvin)
    )
    return np.around(oxy_ml_l, decimals)


def sbe43_hysteresis_voltage(volts, p, coefs, sample_freq=24):
    """
    SBE equation for removing hysteresis from raw voltage values. This function must
    be run before the sbe43 conversion function above.

    Oxygen hysteresis can be corrected after conversion from volts to oxygen
    concentration, see oxy_fitting.hysteresis_correction()

    Parameters
    ----------
    volts : array-like
        Raw voltage
    p : array-like
        CTD pressure values (dbar)
    coefs : dict
        Dictionary of calibration coefficients (H1, H2, H3, offset)
    sample_freq : scalar, optional
        CTD sampling frequency (Hz)

    Returns
    -------
    volts_corrected : array-like
        Hysteresis-corrected voltage

    Notes
    -----
    The hysteresis algorithm is backward-looking so scan 0 must be skipped (as no
    information is available before the first scan).

    See Application Note 64-3 for more information.
    """
    _check_coefs(coefs, ["H1", "H2", "H3", "offset"])
    volts = _check_volts(volts)

    dt = 1 / sample_freq
    D = 1 + coefs["H1"] * (np.exp(np.array(p) / coefs["H2"]) - 1)
    C = np.exp(-1 * dt / coefs["H3"])

    oxy_volts = volts + coefs["offset"]
    oxy_volts_new = np.zeros(oxy_volts.shape)
    oxy_volts_new[0] = oxy_volts[0]
    for i in np.arange(1, len(oxy_volts)):
        oxy_volts_new[i] = (
            (oxy_volts[i] + (oxy_volts_new[i - 1] * C * D[i])) - (oxy_volts[i - 1] * C)
        ) / D[i]

    volts_corrected = oxy_volts_new - coefs["offset"]

    return volts_corrected


def wetlabs_eco_fl(volts, coefs, decimals=4):
    """
    SBE equation for converting ECO-FL fluorometer voltage to concentration.
    SensorID: 20

    Parameters
    ----------
    volts : array-like
        Raw voltage
    coefs : dict
        Dictionary of calibration coefficients (ScaleFactor, DarkOutput/Vblank)

    Returns
    -------
    chl : array-like
        Converted chlorophyll concentration

    Notes
    -----
    Chlorophyll units depend on scale factor (e.g. ug/L-volt, ug/L-counts, ppb/volts),
    see Application Note 62 for more information.
    """
    volts = _check_volts(volts)

    if "DarkOutput" in coefs.keys():
        chl = coefs["ScaleFactor"] * (volts - coefs["DarkOutput"])
    elif "Vblank" in coefs.keys():  # from older calibration sheets
        chl = coefs["ScaleFactor"] * (volts - coefs["Vblank"])
    else:
        log.warning(
            "No dark cast coefficient ('DarkOutput' or 'Vblank'), returning voltage."
        )
        chl = volts

    return np.around(chl, decimals)


def wetlabs_cstar(volts, coefs, decimals=4):
    """
    SBE equation for converting C-Star transmissometer voltage to light transmission.
    SensorID: 71

    Parameters
    ----------
    volts : array-like
        Raw voltage
    coefs : dict
        Dictionary of calibration coefficients (M, B, PathLength)

    Returns
    -------
    xmiss : array-like
        Light transmission [%]
    c : array-like
        Beam attenuation coefficient

    Notes
    -----
    M and B can be recalculated in the field by measuring voltage in air (A1) and
    voltage with the path blocked (Y1):
        M = (Tw / (W0 - Y0)) * (A0 - Y0) / (A1 - Y1)
        B = -M * Y1
    where A0, Y0, and W0 are factory values.
    For transmission relative to water, set Tw = 100%.

    See Application Note 91 for more information.
    """
    _check_coefs(coefs, ["M", "B", "PathLength"])
    volts = _check_volts(volts)
    xmiss = (coefs["M"] * volts) + coefs["B"]  # xmiss as a percentage
    c = -(1 / coefs["PathLength"]) * np.log(xmiss * 100)  # needs xmiss as a decimal

    return np.around(xmiss, decimals), np.around(c, decimals)


def seapoint_fluor(volts, coefs, decimals=6):
    """
    Raw voltage supplied from fluorometer right now, after looking at xmlcon.
    The method will do nothing but spit out the exact values that came in.
    SensorID: 11

    Parameters
    ----------
    volts : array-like
        Raw voltage
    coefs : dict
        Dictionary of calibration coefficients (GainSetting, Offset)

    Returns
    -------
    fluoro : array-like
        Raw voltage

    Notes
    -----
    According to .xmlcon, GainSetting "is an array index, not the actual gain setting."
    """
    _check_coefs(coefs, ["GainSetting", "Offset"])
    volts = np.array(volts)
    fluoro = np.around(volts, decimals)

    return fluoro


def sbe_flntu_chl(volts, coefs, decimals = 4):
    """
    SBE equation for converting SeaBird fluorometer and nepholometric turbidity
    combo sensor's chlorophyll and CDOM fluorometer.
    SensorID: 19
    Example coefs: {'ScaleFactor':11, 'DarkVoltage':0.079}

    Paramters
    ----------
    volts : array-like
        Raw voltage
    coefs : dict
        Dictionary of calibration coefficients (ScaleFactor, Vblank)

    Returns
    -------
    chl : array-like
        Converted chlorophyll concentration in μg/l
    
    Notes:
    ------
    ScaleFactor is usually given in volts, whereas Vblank are in μg/l/V in
    analogue calculations.
    """
    _check_coefs(coefs, ["ScaleFactor", "Vblank"])
    volts = _check_volts(volts)
    chl = np.around((volts - coefs["Vblank"]) * coefs["ScaleFactor"], decimals)

    return chl


def sbe_flntu_ntu(volts, coefs, decimals = 4): 
    """
    SBE equation for converting SeaBird fluorometer and nepholometric turbidity
    combo sensor's turbidity channel.
    SensorID: 67
    Example coefs: {'ScaleFactor':5, 'DarkVoltage':0.05}

    Paramters
    ----------
    volts : array-like
        Raw voltage
    coefs : dict
        Dictionary of calibration coefficients (ScaleFactor, DarkVoltage)

    Returns
    -------
    turb : array-like
        Converted turbidity (units expressed in NTU)
    
    Notes:
    ------
    ScaleFactor is usually given in volts, whereas DarkVoltage are in NTU/V in
    analogue calculations.
    Future release may factor this into sbe_flntu_chl (same equation)
    """

    _check_coefs(coefs, ["ScaleFactor", "DarkVoltage"])
    volts = _check_volts(volts)
    turb = np.around((volts - coefs["DarkVoltage"]) * coefs["ScaleFactor"], decimals)

    return turb



import numpy as np
import pandas as pd
import pytest

from ctdcal import equations_sbe as eqs


def make_coefs(*args, value=0):
    return {x: value for x in args[0]}


def test_sbe3(caplog):
    freq = 99 * [0] + [1]
    coefs = ["G", "H", "I", "J", "F0"]

    # check values are converted correctly
    cnv = eqs.sbe3(freq, make_coefs(coefs, value=1))  # avoid divide by zero
    assert "int" in caplog.records[0].message  # check logging output
    assert "sbe3" in caplog.records[1].message  # check logging output
    assert all(np.isnan(cnv[:-1]))
    assert cnv[-1] == -272.15  # t_ITS90 = (1 / 1) - 273.15
    assert cnv.dtype == float

    # check there are no warnings if freq is float with no zeroes
    no_warn = eqs.sbe3(np.ones(len(freq)), make_coefs(coefs, value=1))
    assert len(caplog.records) == 2
    assert all(no_warn == -272.15)
    assert no_warn.dtype == float

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="F0"):
        eqs.sbe3(freq, make_coefs(coefs[:-1]))


def test_sbe4(caplog):
    freq = 99 * [0] + [1]
    t = len(freq) * [0]
    p = len(freq) * [0]
    coefs = ["G", "H", "I", "J", "CPcor", "CTcor"]

    # check values are converted correctly
    cnv = eqs.sbe4(freq, t, p, make_coefs(coefs, value=1))  # avoid divide by zero
    assert "int" in caplog.records[0].message  # check logging output
    assert "sbe4" in caplog.records[1].message  # check logging output
    assert all(np.isnan(cnv[:-1]))
    assert cnv[-1] == 1
    assert cnv.dtype == float

    # check there are no warnings if freq is float with no zeroes
    no_warn = eqs.sbe4(np.ones(len(freq)), t, p, make_coefs(coefs, value=1))
    assert len(caplog.records) == 2
    assert all(no_warn == 1)
    assert no_warn.dtype == float

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="CTcor"):
        eqs.sbe4(freq, t, p, make_coefs(coefs[:-1]))


def test_sbe9(caplog):
    freq = 99 * [0] + [1]
    t = len(freq) * [0]
    coefs = (
        ["T1", "T2", "T3", "T4", "T5"]
        + ["C1", "C2", "C3"]
        + ["D1", "D2"]
        + ["AD590M", "AD590B"]
    )

    # check values are converted correctly
    cnv = eqs.sbe9(freq, t, make_coefs(coefs))
    assert "int" in caplog.records[0].message  # check logging output
    assert "sbe9" in caplog.records[1].message  # check logging output
    assert all(np.isnan(cnv[:-1]))
    assert cnv[-1] == -10.1353
    assert cnv.dtype == float

    # check there are no warnings if freq is float with no zeroes
    no_warn = eqs.sbe9(np.ones(len(freq)), t, make_coefs(coefs))
    assert len(caplog.records) == 2
    assert all(no_warn == -10.1353)
    assert no_warn.dtype == float

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="AD590B"):
        eqs.sbe9(freq, t, make_coefs(coefs[:-1]))


def test_sbe_altimeter(caplog):
    volts = 99 * [0] + [1]
    coefs = ["ScaleFactor", "Offset"]

    # check values are converted correctly
    cnv = eqs.sbe_altimeter(volts, make_coefs(coefs, value=1))  # avoid divide by zero
    assert "int" in caplog.records[0].message
    assert "sbe_altimeter" in caplog.records[0].message
    assert all(cnv[:-1] == 1)
    assert cnv[-1] == 301
    assert cnv.dtype == float

    # check there are no warnings if volts are float with no zeroes
    no_warn = eqs.sbe_altimeter(np.ones(len(volts)), make_coefs(coefs, value=1))
    assert len(caplog.records) == 1
    assert all(no_warn == 301)
    assert no_warn.dtype == float

    # check voltages outside 0-5V are replaced with NaNs
    bad_range = eqs.sbe_altimeter(np.arange(-1, 7), make_coefs(coefs, value=1))
    assert "sbe_altimeter" in caplog.records[-1].message
    assert "outside of 0-5V" in caplog.records[-1].message
    assert np.isnan(bad_range[0])
    assert np.isnan(bad_range[-1])
    assert not all(np.isnan(bad_range[1:-1]))

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="Offset"):
        eqs.sbe_altimeter(volts, make_coefs(coefs[:-1]))


def test_sbe43(caplog):
    volts = 99 * [0] + [1]
    p = len(volts) * [0]
    t = len(volts) * [0]
    c = len(volts) * [0]
    coefs = ["Soc", "offset", "Tau20", "A", "B", "C", "E"]

    # check values are converted correctly
    cnv = eqs.sbe43(volts, p, t, c, make_coefs(coefs, value=1))
    assert "int" in caplog.records[0].message
    assert "sbe43" in caplog.records[0].message
    assert all(cnv[:-1] == 10.2314)
    assert cnv[-1] == 20.4628
    assert cnv.dtype == float

    # check there are no warnings if volts are float with no zeroes
    no_warn = eqs.sbe43(np.ones(len(volts)), p, t, c, make_coefs(coefs, value=1))
    assert len(caplog.records) == 1
    assert all(no_warn == 20.4628)
    assert no_warn.dtype == float

    # check voltages outside 0-5V are replaced with NaNs
    volts = np.arange(-1, 7)
    p, t, c = tuple(np.ones(len(volts)) for x in [1, 2, 3])  # definition shorthand
    bad_range = eqs.sbe43(volts, p, t, c, make_coefs(coefs, value=1))
    assert "sbe43" in caplog.records[-1].message
    assert "outside of 0-5V" in caplog.records[-1].message
    assert np.isnan(bad_range[0])
    assert np.isnan(bad_range[-1])
    assert not all(np.isnan(bad_range[1:-1]))

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="E"):
        eqs.sbe43(volts, p, t, c, make_coefs(coefs[:-1]))


def test_sbe43_hysteresis_voltage(caplog):
    volts = 99 * [0] + [1]
    p = len(volts) * [0]
    coefs = ["H1", "H2", "H3", "offset"]

    # check values are converted correctly
    cnv = eqs.sbe43_hysteresis_voltage(volts, p, make_coefs(coefs, value=1))
    assert "int" in caplog.records[0].message
    assert "sbe43_hysteresis_voltage" in caplog.records[0].message
    assert all(cnv[:-1] == 0)
    assert cnv[-1] == 1
    assert cnv.dtype == float

    # check there are no warnings if volts are float with no zeroes
    no_warn = eqs.sbe43_hysteresis_voltage(
        np.ones(len(volts)), p, make_coefs(coefs, value=1)
    )
    assert len(caplog.records) == 1
    assert all(no_warn == 1)
    assert no_warn.dtype == float

    # check voltages outside 0-5V are replaced with NaNs
    volts = np.arange(0, 7)
    p = np.ones(len(volts))
    bad_range = eqs.sbe43_hysteresis_voltage(volts, p, make_coefs(coefs, value=1))
    assert "sbe43_hysteresis_voltage" in caplog.records[-1].message
    assert "outside of 0-5V" in caplog.records[-1].message
    assert not all(np.isnan(bad_range[:-1]))
    assert np.isnan(bad_range[-1])

    # check NaNs are propagated through hysteresis correction
    volts = np.arange(-1, 7)
    p = np.ones(len(volts))
    bad_range = eqs.sbe43_hysteresis_voltage(volts, p, make_coefs(coefs, value=1))
    assert all(np.isnan(bad_range))

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="offset"):
        eqs.sbe43_hysteresis_voltage(volts, p, make_coefs(coefs[:-1]))


def test_wetlabs_eco_fl(caplog):
    volts = 99 * [0] + [1]
    coefs = ["ScaleFactor", "DarkOutput"]

    # check values are converted correctly
    cnv = eqs.wetlabs_eco_fl(volts, make_coefs(coefs, value=1))
    assert "int" in caplog.records[0].message
    assert "wetlabs_eco_fl" in caplog.records[0].message
    assert all(cnv[:-1] == -1)
    assert cnv[-1] == 0
    assert cnv.dtype == float

    # check for same result with "Vblank"
    coefs = ["ScaleFactor", "Vblank"]
    cnv_Vblank = eqs.wetlabs_eco_fl(volts, make_coefs(coefs, value=1))
    assert all(cnv == cnv_Vblank)

    # check there are no warnings if volts are float with no zeroes
    no_warn = eqs.wetlabs_eco_fl(np.ones(len(volts)), make_coefs(coefs, value=1))
    assert len(caplog.records) == 2
    assert all(no_warn == 0)
    assert no_warn.dtype == float

    # check voltages outside 0-5V are replaced with NaNs
    bad_range = eqs.wetlabs_eco_fl(np.arange(-1, 7), make_coefs(coefs, value=1))
    assert "wetlabs_eco_fl" in caplog.records[-1].message
    assert "outside of 0-5V" in caplog.records[-1].message
    assert np.isnan(bad_range[0])
    assert np.isnan(bad_range[-1])
    assert not all(np.isnan(bad_range[1:-1]))

    # return volts if "DarkOutput" or "Vblank" not in coefs
    no_cnv = eqs.wetlabs_eco_fl(np.ones(len(volts)), make_coefs(["ScaleFactor"]))
    assert "returning voltage" in caplog.records[-1].message
    assert all(no_cnv == np.ones(len(volts)))


def test_wetlabs_cstar(caplog):
    volts = 99 * [0] + [1]
    coefs = ["M", "B", "PathLength"]

    # check values are converted correctly
    cnv = eqs.wetlabs_cstar(volts, make_coefs(coefs, value=1))  # avoid divide by zero
    xmiss, c = cnv
    assert "int" in caplog.records[0].message
    assert "wetlabs_cstar" in caplog.records[0].message
    assert all(xmiss[:-1] == 1)
    assert xmiss[-1] == 2
    assert all(c[:-1] == -4.6052)
    assert c[-1] == -5.2983

    # check there are no warnings if volts are float with no zeroes
    no_warn = eqs.wetlabs_cstar(np.ones(len(volts)), make_coefs(coefs, value=1))
    assert len(caplog.records) == 1
    assert all(no_warn[0] == 2)
    assert all(no_warn[1] == -5.2983)

    # check voltages outside 0-5V are replaced with NaNs
    bad_range = eqs.wetlabs_cstar(np.arange(-1, 7), make_coefs(coefs, value=1))
    xmiss, c = bad_range
    assert "wetlabs_cstar" in caplog.records[-1].message
    assert "outside of 0-5V" in caplog.records[-1].message
    assert np.isnan(xmiss[0])
    assert np.isnan(xmiss[-1])
    assert not all(np.isnan(xmiss[1:-1]))
    assert np.isnan(c[0])
    assert np.isnan(c[-1])
    assert not all(np.isnan(c[1:-1]))

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="PathLength"):
        eqs.wetlabs_cstar(volts, make_coefs(coefs[:-1]))


def test_seapoint_fluor():
    volts = 99 * [0] + [1]
    coefs = ["GainSetting", "Offset"]

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="Offset"):
        eqs.seapoint_fluor(volts, make_coefs(coefs[:-1]))

def test_sbe_flntu_chl(caplog):
    volts = np.concatenate((np.zeros(10), np.array([1, 2, 3, 4, 5])))
    coefs = {'ScaleFactor':11, 'Vblank':0.079}

    chl = eqs.sbe_flntu_chl(volts, coefs)

    #   Test hand-calculated values with example coefs
    assert all(chl[0:10] == -0.869)
    assert chl[-1] == 54.131 

    # check there are no warnings if volts are float with no zeroes
    no_warn = eqs.sbe_flntu_chl(np.ones(len(volts)), make_coefs(coefs, value=1))
    assert len(caplog.records) == 0
    assert all(no_warn == 0)

    # check voltages outside 0-5V are replaced with NaNs
    bad_range = eqs.sbe_flntu_chl(np.arange(-1, 7), make_coefs(coefs, value=1))
    assert "sbe_flntu_chl" in caplog.records[-1].message
    assert "outside of 0-5V" in caplog.records[-1].message
    assert np.isnan(bad_range[0])
    assert np.isnan(bad_range[-1])
    assert bad_range[1] == -1
    assert not all(np.isnan(bad_range[1:-1]))

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="Vblank"):
        eqs.sbe_flntu_chl(volts, {"ScaleFactor":1, "dark_counts":1})
        assert "dictionary missing keys" in caplog.records[-1].message

def test_sbe_flntu_ntu(caplog):
    #   Practically identical to sbe_flntu_chl
    volts = np.concatenate((np.zeros(10), np.array([1, 2, 3, 4, 5])))
    coefs = {'ScaleFactor':5, 'DarkVoltage':0.050}

    ntu = eqs.sbe_flntu_ntu(volts, coefs)

    #   Test hand-calculated values with example coefs
    assert all(ntu[0:10] == -0.25)
    assert ntu[-1] == 24.75

    # check there are no warnings if volts are float with no zeroes
    no_warn = eqs.sbe_flntu_ntu(np.ones(len(volts)), make_coefs(coefs, value=1))
    assert len(caplog.records) == 0
    assert all(no_warn == 0)

    # check voltages outside 0-5V are replaced with NaNs
    bad_range = eqs.sbe_flntu_ntu(np.arange(-1, 7), make_coefs(coefs, value=1))
    assert "sbe_flntu_ntu" in caplog.records[-1].message
    assert "outside of 0-5V" in caplog.records[-1].message
    assert np.isnan(bad_range[0])
    assert np.isnan(bad_range[-1])
    assert bad_range[1] == -1
    assert not all(np.isnan(bad_range[1:-1]))

    # error saying which keys are missing from coef dict
    with pytest.raises(KeyError, match="DarkVoltage"):
        eqs.sbe_flntu_ntu(volts, {"ScaleFactor":1, "dark_counts":1})
        assert "dictionary missing keys" in caplog.records[-1].message


"""
Tools for cleaning up converted and processed cast data.
"""
import logging
from pathlib import Path

import pandas as pd
import numpy as np
from scipy import signal as sig


log = logging.getLogger(__name__)


class Cast(object):
    """
    Cast data container with methods for separating upcast, filtering
    and trimming soak period.

    Attributes
    ----------
    cast_id : str
        Cast identifier.
    datadir : str or Path-like
        Data directory.
    p_col : str
        Name of pressure column.
    proc : DataFrame
        Pre-processed cast data.
    downcast : DataFrame
        Downast data trimmed from full cast data.
    filtered : array-like
        Filtered cast data.
    trimmed : DataFrame
        Upcast data trimmed of initial soak data.
    ondeck_trimmed : DataFrame
        Full cast data trimmed of on-deck intervals.
    """
    def __init__(self, cast_id, datadir):
        self.cast_id = cast_id
        self.datadir = datadir
        self.p_col = None
        self.proc = None
        self.downcast = None
        self.filtered = None
        self.trimmed = None
        self.ondeck_trimmed = None
        self.load_cast()

    def load_cast(self):
        """
        Read the processed data into a dataframe.
        """
        f = Path(self.datadir, 'converted/%s.pkl' % self.cast_id)
        self.proc = pd.read_pickle(f)

    def parse_downcast(self, data):
        """
        Trim away the upcast starting from the max value of the pressure
        column.

        Parameters
        ----------
        data : DataFrame
            Cast data.
        """
        if self.p_col is None:
            raise AttributeError("Pressure column 'p_col' attribute is not set")
        downcast = data.loc[:self.proc[self.p_col].idxmax()].copy()
        self.downcast = downcast

    def filter(self, data, win_size, win_type="triangle", cols=None):
        """
        Filter processed CTD data using one of three window types (boxcar,
        hann, triangle).

        Previously was 'raw_ctd_filter' in ctdcal.process_ctd.

        Parameters
        ----------
        data : array-like
            Cast data.
        win_size : int
            Window size in number of samples.
        win_type : str
            Window type.
        cols : list
            Column names to filter.
        """
        data = data.copy()
        if win_type == "boxcar":
            win = sig.windows.boxcar(win_size)
        elif win_type == "hann":
            win = sig.windows.hann(win_size)
        elif win_type == "triangle":
            win = sig.windows.triang(win_size)
        else:
            raise AttributeError('Error filtering cast %s! No filter of type %s found.' % (self.cast_id, win_type))
        for col in cols:
            data[col] = sig.convolve(data[col], win, mode="same") / np.sum(win)
        self.filtered = data

    def trim_soak(self, data, win_size, max_soak=20):
        """
        Detects the end of the soak period by identifying local peaks in the
        cast data, and choosing the last one above the threshold.

        data : DataFrame
            Downcast data.
        win_size : int
            Window size in number of samples.
        max_soak : int
            Soak threshold in cast pressure units.
        """
        if self.p_col is None:
            raise AttributeError("Pressure column 'p_col' attribute is not set")
        if self.downcast is None:
            raise AttributeError("'downcast' attribute is not set.")
        if win_size < 0:
            win_size = 0  # must be positive or zero

        # Adapted from: https://stackoverflow.com/questions/48023982/pandas-finding-local-max-and-min
        # Find local peaks
        local_min_vals = data.loc[data[self.p_col] == data[self.p_col].rolling(win_size, center=True).min(),
                                  [self.p_col]]

        # apply max_soak threshold to filter false positives...
        i = local_min_vals.index[-1]
        while local_min_vals.loc[i, self.p_col] > max_soak:
            try:
                local_min_vals.drop(i, inplace=True)
            except KeyError:
                log.warning('Whoa, trouble finding the soak on cast %s! Nothing was trimmed!' % self.cast_id)
                return
            i = local_min_vals.index[-1]

        soak_end = local_min_vals.index[-1]
        self.trimmed = self.downcast.loc[soak_end:]

    def get_details(self):
        """
        Collect cast details for export (see Notes).

        Previously was 'cast_details' in ctdcal.process_ctd.

        Notes
        -----
        The following variables are returned in an array:
        cast_id : cast identifier
        start_time : Time at start of cast (in unix epoch time)
        end_time : Time at end of cast (in unix epoch time)
        bottom_time : Time at bottom of cast (in unix epoch time)
        start_pressure : Pressure at which cast started
        max_pressure : Bottom of the cast pressure
        latitude : Latitude at bottom of cast
        longitude : Longitude at bottom of cast
        altimeter_bottom : Altimeter reading at bottom of cast

        Returns
        -------
        DataFrame
        """
        if self.p_col is None:
            raise AttributeError("Pressure column 'p_col' attribute is not set")
        if self.trimmed is None:
            raise AttributeError("'trimmed' attribute is not set.")
        data = pd.DataFrame()
        data['cast_id'] = [self.cast_id]
        best_full_cast = self.filtered if self.filtered is not None else self.proc
        # TODO: the below uses all very odf/go-ship/seabird-specific column labels
        #   which should ultimately be replaced with standardized names.
        data['start_time'] = [float(self.trimmed["scan_datetime"].head(1))]
        p_max_ind = best_full_cast[self.p_col].argmax()
        data['bottom_time'] = [float(best_full_cast["scan_datetime"][p_max_ind])]
        data['end_time'] = [float(best_full_cast["scan_datetime"].tail(1))]
        data['start_pressure'] = [float(np.around(self.trimmed[self.p_col].head(1), 4))]
        data['max_pressure'] = [float(np.around(self.trimmed[self.p_col].max(), 4))]
        if 'ALT' in self.proc.columns:
            data['altimeter_bottom'] = [float(np.around(best_full_cast["ALT"][p_max_ind], 4))]
        if all(col in self.proc.columns for col in ['GPSLAT', 'GPSLON']):
            data['latitude'] = [float(np.around(best_full_cast["GPSLAT"][p_max_ind], 4))]
            data['longitude'] = [float(np.around(best_full_cast["GPSLON"][p_max_ind], 4))]
        return data

    def get_pressure_offsets(self, data, threshold, sample_freq):
        """
        Get starting and ending on-deck pressure averages for the cast. A
        conductivity threshold is used to determine when the instrument
        is in or out of the water. A default conductivity threshold is
        set in the user configuration file.

        Previously was 'remove_on_deck' in ctdcal.process_ctd.

        Parameters
        ----------
        data : DataFrame
            Cast data.
        threshold : float
            Maximum conductivity value to determine when cast is not in the water,
            in source conductivity units.
        sample_freq : int
            Instrument sample frequency in Hz

        Returns
        -------
        DataFrame
        """
        # TODO: review algorithm for soundness and reliability, see if we can
        #     use pressure or something other than salinity, remove reliance
        #     on WOCE colnames

        # Half minute
        time_delay = sample_freq * 30  # time to let CTD pressure reading settle/sit on deck
        # split dataframe into upcast/downcast
        downcast = data.iloc[: (data[self.p_col].argmax() + 1)]
        upcast = data.iloc[(data[self.p_col].argmax() + 1):]
        # Search each half of df for minimum conductivity
        # threshold to identify when rosette is out of water
        start_df = downcast.loc[
            (downcast['CTDCOND1'] < threshold)
            & (downcast['CTDCOND2'] < threshold),
            self.p_col,
        ]
        end_df = upcast.loc[
            (upcast['CTDCOND1'] < threshold)
            & (upcast['CTDCOND2'] < threshold),
            self.p_col,
        ]
        # Evaluate starting and ending pressures
        start_samples = len(start_df)
        if start_samples > time_delay:
            start_p = np.average(start_df.iloc[(sample_freq * 2): (start_samples - time_delay)])
        else:
            start_seconds = start_samples / sample_freq
            log.warning(
                    f"{self.cast_id}: Only {start_seconds:0.1f} seconds of start pressure averaged."
            )
            start_p = np.average(start_df.iloc[(sample_freq * 2):start_samples])
        end_samples = len(end_df)
        if end_samples > time_delay:
            end_p = np.average(end_df.iloc[time_delay:])
        else:
            end_seconds = end_samples / sample_freq
            log.warning(
                    f"{self.cast_id}: Only {end_seconds:0.1f} seconds of end pressure averaged."
            )
            end_p = np.average(end_df)  # just average whatever there is
        # Warn if failures
        if len(start_df) == 0:
            log.warning("Failed to find starting deck pressure.")
        if len(end_df) == 0:
            log.warning("Failed to find ending deck pressure.")

        df = pd.DataFrame()
        df['cast_id'] = [self.cast_id]
        df['pressure_start'] = [start_p]
        df['pressure_end'] = [end_p]
        return df

    def trim_ondeck(self, data, threshold):
        """
        Trims end of casts based on a conductivity threshold. Legacy functionality.

        Parameters
        ----------
        data : DataFrame
            Cast data.
        threshold : int
            Maximum conductivity value to determine when cast is not in the water,
            in source conductivity units.
        """
        # split dataframe into upcast/downcast
        # split dataframe into upcast/downcast
        downcast = data.iloc[: (data[self.p_col].argmax() + 1)]
        upcast = data.iloc[(data[self.p_col].argmax() + 1):]
        # Search each half of df for minimum conductivity
        # threshold to identify when rosette is out of water
        start_df = downcast.loc[
            (downcast['CTDCOND1'] < threshold)
            & (downcast['CTDCOND2'] < threshold),
            self.p_col,
        ]
        end_df = upcast.loc[
            (upcast['CTDCOND1'] < threshold)
            & (upcast['CTDCOND2'] < threshold),
            self.p_col,
        ]
        self.ondeck_trimmed = data.iloc[start_df.index.max() : end_df.index.min()].copy()



"""External lookup file for sensors, header information.
This file may need to be edited at sea if a sensor is attached to a package that
this software has not encountered before. In that case, follow the steps below:

1. On the SBE acquisition machine, open the .XMLCON for a cast.
2. Look down the .XMLCON file until you find a line with:
        <Sensor index="xx" SensorID="yy" >
   Where "xx" and "yy" are numbers, and the following lines are the sensor to edit in.
3. In the short_lookup table below, copy and paste a line, then edit the line:
        'SensorID':{'short_name': 'mm', 'long_name':'ll', 'units': 'uu', 'type': 'tt'}
    Where mm is a science defined short name, ll is the name of the sensor,
    uu is the unit used by the sensor, and tt is the data, computer format of the sensor,
    which will typically be float64. (Any sensor that puts out a decimal number is float64)
4. If adding at the end of the short_lookup table, remember to add a comma ',' to the
    previous line, if adding at the top or middle add a comma to the inserted line.
5. Save the file, and attempt to rerun the program again.

Questions for the future:
-This is SBE focused right now. What do we do for potential namespace conflicts with
other CTDs/sensors from other companies (RBR, others)?
-CTDs send up signals and could have varying levels of precision. Do we put a default
truncation value in a lookup table, or make sure the corresponding equation has it built in
(which again with SBE CTDs are dependent on how the CTD logs it, unless it has a datalogger),
or a third option?

Joseph Gum
November 10, 2016
"""

#lookup table for sensor data
###DOUBLE CHECK TYPE IS CORRECT###
short_lookup = {
    '55':{'short_name': 't', 'long_name':'SBE 3+ Temperature', 'units': 'C', 'type': 'float64'},
    '45':{'short_name': 'p', 'long_name':'SBE 9+ Pressure', 'units': 'dbar', 'type': 'float64'},
    '3':{'short_name': 'c', 'long_name':'SBE 4 Conductivity', 'units': 'S/m', 'type':'float64'},
    '38':{'short_name': 'o', 'long_name':'SBE 43 Oxygen', 'units': 'ml/l', 'type':'float64'},
    '11':{'short_name': 'fluoro', 'long_name':'Seapoint Fluorometer', 'units': 'ug/l', 'type':'float64'},
    '27':{'short_name': 'empty', 'long_name':'empty', 'units':'NA', 'type':'NA'},
    '0':{'short_name': 'alti', 'long_name':'Altitude', 'units':'m', 'type':'float64'},
    '71':{'short_name': 'cstar', 'long_name':'CStar Transmissometer', 'units': 'ug/l', 'type':'float64'},
    '61':{'short_name': 'u_def', 'long_name':'user defined', 'units':'V', 'type':'float64'},
    '1000':{'short_name': 'sal', 'long_name':'Salinity (C1 T1)', 'units':'PSU', 'type':'float64'}
}

#because it's easier to not think on long flights and instead write mindless code.
scan_header_lookup = {
    'index':{'long_name': 'Scan Index Number', 'type':'int32'},
    't1_C':{'long_name': 'Temperature 1', 'type':'float64'},
    't2_C':{'long_name': 'Temperature 2', 'type':'float64'},
    'p_dbar':{'long_name': 'Pressure', 'type':'float64'},
    'c1_mS/cm':{'long_name': 'Conductivity 1', 'type':'float64'},
    'c2_mS/cm':{'long_name': 'Conductivity 2', 'type':'float64'},
    'sal_PSU':{'long_name': 'Salinity (T1C1)', 'type':'float64'},
    'o1_ml/l':{'long_name': 'Oxygen 1', 'type':'float64'},
    'empty1_NA':{'long_name': 'Empty', 'type':'float64'},
    'empty2_NA':{'long_name': 'Empty', 'type':'float64'},
    'u_def1_V':{'long_name': 'User Defined 1', 'type':'float64'},
    'u_def2_V':{'long_name': 'User Defined 2', 'type':'float64'},
    'fluoro_ug/l':{'long_name': 'Fluorometer', 'type':'float64'},
    'alti_m':{'long_name': 'Altimeter', 'type':'float64'},
    'cstar_ug/l'{'long_name': 'CStar Transmissometer', 'type':'float64'},
    'lat_ddeg':{'long_name':'Latitude Decimal Degrees', 'type':'float64'},
    'lon_ddeg': {'long_name':'Longitude Decimal Degrees', 'type':'float64'},
    'new_fix_bool': {'long_name':'New Fix?', 'type':'bool_'},
    'nmea_datetime': {'long_name':'NMEA Date/Time', 'type':'datetime64'},
    'scan_datetime': {'long_name':'Scan Date/Time', 'type':'datetime64'},
    'btl_fire_bool': {'long_name':'Bottle Fire', 'type':'bool_'}
}



from pathlib import Path
from unittest.mock import patch

import pytest
import yaml

from ctdcal.common import load_user_config, validate_dir, validate_file


class TestUserConfig:
    """
    Tests and fixtures for working with user configurations.
    """
    @pytest.fixture
    def tmp_cfg(self, tmp_path):
        fname = tmp_path / 'cfg.yaml'
        with open(fname, 'w') as f:
            yaml.dump({'cheese': {'cheddar': 0, 'limburger': 1}}, f)
        return fname

    def test_load_user_config(self, tmp_cfg):
        # Test good cfgfile
        cfg = load_user_config(tmp_cfg)
        assert cfg.cheese.limburger == 1
        # Test bad cfgfile
        # TODO: Test for invalid file?


class TestPathValidation:
    """
    Unit tests and fixtures for file and directory-related validation
    functions.
    """
    @pytest.fixture
    def tmp_dir(self, tmp_path):
        path = tmp_path
        return path

    def test_validate_dir(self, tmp_dir):
        p = 'spam/'
        # Test for bad input
        with pytest.raises(TypeError):
            validate_dir(None)

        # Test for bad path
        with pytest.raises(FileNotFoundError):
            validate_dir(p, create=False)

        # Test for good path
        with patch('pathlib.Path.is_dir', return_value=True):
            validated_path = validate_dir(p, create=False)
        assert validated_path == Path(p)

        # Test that bad dir is created
        test_dir = tmp_dir / p / 'egg'
        assert test_dir.is_dir() is False
        assert validate_dir(test_dir, create=True).is_dir()

        # Test for path exists but is not a dir
        samename = tmp_dir / p / 'cheese'
        assert samename.exists() is False
        samename.touch()
        with pytest.raises(FileExistsError):
            validate_dir(samename, create=False)
        with pytest.raises(FileExistsError):
            validate_dir(samename, create=True)

    def test_validate_file(self, tmp_dir):
        p = 'spam.file'
        # Test for bad input
        with pytest.raises(TypeError):
            validate_file(None)

        # Test for bad path
        with pytest.raises(FileNotFoundError):
            validate_file(p, create=False)

        # Test for good path
        with patch('pathlib.Path.is_file', return_value=True):
            validated_path = validate_file(p, create=False)
        assert validated_path == Path(p)

        # Test that non-existent file is created
        tmp_file = tmp_dir / 'spam' / 'egg.file'
        assert tmp_file.is_file() is False
        assert validate_file(tmp_file, create=True).is_file()

        # Test that filename exists but is not a file
        samename = tmp_dir / 'spam' / 'egg'
        assert samename.exists() is False
        samename.mkdir(parents=True)
        with pytest.raises(FileExistsError):
            validate_file(samename, create=False)
        with pytest.raises(FileExistsError):
            validate_file(samename, create=True)
